"""Advance Analytics for Engagements"""
from datetime import datetime

from edx_rbac.decorators import permission_required
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from rest_framework import status as rest_status
from rest_framework.response import Response
from rest_framework.views import APIView

from django.http import HttpResponse, StreamingHttpResponse

from enterprise_data.admin_analytics.constants import Calculation, EngagementChart, Granularity, ResponseType
from enterprise_data.admin_analytics.utils import (
    calculation_aggregation,
    fetch_and_cache_engagements_data,
    fetch_and_cache_enrollments_data,
    fetch_enrollments_cache_expiry_timestamp,
    granularity_aggregation,
)
from enterprise_data.api.v1 import serializers
from enterprise_data.api.v1.paginators import AdvanceAnalyticsPagination
from enterprise_data.renderers import IndividualEngagementsCSVRenderer
from enterprise_data.utils import date_filter


class AdvanceAnalyticsIndividualEngagementsView(APIView):
    """
    API for getting the advance analytics individual engagements data.
    """

    authentication_classes = (JwtAuthentication,)
    pagination_class = AdvanceAnalyticsPagination
    http_method_names = ["get"]

    @permission_required('can_access_enterprise', fn=lambda request, enterprise_uuid: enterprise_uuid)
    def get(self, request, enterprise_uuid):
        """
        HTTP GET endpoint to retrieve the enterprise engagements data.
        """
        serializer = serializers.AdvanceAnalyticsEngagementStatsSerializer(data=request.GET)
        serializer.is_valid(raise_exception=True)
        cache_expiry = fetch_enrollments_cache_expiry_timestamp()

        enrollment_df = fetch_and_cache_enrollments_data(enterprise_uuid, cache_expiry).copy()
        engagement_df = fetch_and_cache_engagements_data(enterprise_uuid, cache_expiry).copy()
        # Use start and end date if provided by the client, if client has not provided then use
        # 1. minimum enrollment date from the data as the start_date
        # 2. today's date as the end_date
        start_date = serializer.data.get('start_date', enrollment_df.enterprise_enrollment_date.min())
        end_date = serializer.data.get('end_date', datetime.now())
        response_type = request.query_params.get('response_type', ResponseType.JSON.value)
        # Date filtering.
        engagements = date_filter(
            start=start_date, end=end_date, data_frame=engagement_df.copy(), date_column='activity_date'
        )
        engagements["learning_time_hours"] = engagements["learning_time_seconds"] / 60 / 60
        engagements = engagements[engagements["learning_time_hours"] > 0]
        engagements["learning_time_hours"] = round(engagements["learning_time_hours"].astype(float), 1)

        # Select only the columns that will be in the table.
        engagements = engagements[
            [
                "email",
                "course_title",
                "activity_date",
                "course_subject",
                "learning_time_hours",
            ]
        ]
        engagements["activity_date"] = engagements["activity_date"].dt.date
        engagements = engagements.sort_values(by="activity_date", ascending=False).reset_index(drop=True)
        if response_type == ResponseType.CSV.value:
            response = StreamingHttpResponse(
                IndividualEngagementsCSVRenderer().render(self._stream_serialized_data(engagements)),
                content_type="text/csv"
            )
            start_date = start_date.strftime('%Y/%m/%d')
            end_date = end_date.strftime('%Y/%m/%d')
            filename = f"""Individual Engagements, {start_date} - {end_date}.csv"""
            response['Content-Disposition'] = f'attachment; filename="{filename}"'
            response['Access-Control-Expose-Headers'] = 'Content-Disposition'
            return response

        paginator = self.pagination_class()
        page = paginator.paginate_queryset(engagements, request)
        serialized_data = page.data.to_dict(orient='records')
        response = paginator.get_paginated_response(serialized_data)

        return response

    def _stream_serialized_data(self, engagements, chunk_size=50000):
        """
        Stream the serialized data.
        """
        total_rows = engagements.shape[0]
        for start_index in range(0, total_rows, chunk_size):
            end_index = min(start_index + chunk_size, total_rows)
            chunk = engagements.iloc[start_index:end_index]
            yield from chunk.to_dict(orient='records')


class AdvanceAnalyticsEngagementStatsView(APIView):
    """
    API for getting the advance analytics engagements statistics data.
    """

    authentication_classes = (JwtAuthentication,)
    http_method_names = ["get"]

    @permission_required('can_access_enterprise', fn=lambda request, enterprise_uuid: enterprise_uuid)
    def get(self, request, enterprise_uuid):
        """
        HTTP GET endpoint to retrieve the enterprise engagements statistics data.
        """
        serializer = serializers.AdvanceAnalyticsEngagementStatsSerializer(data=request.GET)
        serializer.is_valid(raise_exception=True)

        cache_expiry = fetch_enrollments_cache_expiry_timestamp()

        enrollment_df = fetch_and_cache_enrollments_data(enterprise_uuid, cache_expiry).copy()
        engagement_df = fetch_and_cache_engagements_data(enterprise_uuid, cache_expiry).copy()
        # Use start and end date if provided by the client, if client has not provided then use
        # 1. minimum enrollment date from the data as the start_date
        # 2. today's date as the end_date
        start_date = serializer.data.get('start_date', enrollment_df.enterprise_enrollment_date.min())
        end_date = serializer.data.get('end_date', datetime.now())
        granularity = serializer.data.get('granularity', Granularity.DAILY.value)
        calculation = serializer.data.get('calculation', Calculation.TOTAL.value)
        chart_type = serializer.data.get('chart_type')

        if chart_type is None:
            data = {
                "engagements_over_time": self.construct_engagements_over_time(
                    engagement_df.copy(),
                    start_date,
                    end_date,
                    granularity,
                    calculation,
                ),
                "top_courses_by_engagement": self.construct_top_courses_by_engagements(
                    engagement_df.copy(),
                    start_date,
                    end_date,
                ),
                "top_subjects_by_engagement": self.construct_top_subjects_by_engagements(
                    engagement_df.copy(),
                    start_date,
                    end_date,
                ),
            }
            return Response(data)
        elif chart_type == EngagementChart.ENGAGEMENTS_OVER_TIME.value:
            return self.construct_engagements_over_time_csv(
                engagement_df.copy(),
                start_date,
                end_date,
                granularity,
                calculation,
            )
        elif chart_type == EngagementChart.TOP_COURSES_BY_ENGAGEMENTS.value:
            return self.construct_top_courses_by_engagements_csv(
                engagement_df.copy(),
                start_date,
                end_date,
            )
        elif chart_type == EngagementChart.TOP_SUBJECTS_BY_ENGAGEMENTS.value:
            return self.construct_top_subjects_by_engagements_csv(
                engagement_df.copy(),
                start_date,
                end_date,
            )
        return Response(data='Not Found', status=rest_status.HTTP_400_BAD_REQUEST)

    def engagements_over_time_common(self, engagements_df, start_date, end_date, granularity, calculation):
        """
        Common method for constructing engagements over time data.

        Arguments:
            engagements_df {DataFrame} -- DataFrame of engagements
            start_date {datetime} -- Engagement start date in the format 'YYYY-MM-DD'
            end_date {datetime} -- Engagement end date in the format 'YYYY-MM-DD'
            granularity {str} -- Granularity of the data. One of Granularity choices
            calculation {str} -- Calculation of the data. One of Calculation choices
        """
        engagements_df["learning_time_hours"] = engagements_df["learning_time_seconds"] / 60 / 60
        engagements_df = engagements_df[engagements_df["learning_time_hours"] > 0]
        engagements_df["learning_time_hours"] = round(engagements_df["learning_time_hours"].astype(float), 1)

        engagements_df = engagements_df[["activity_date", "enroll_type", "learning_time_hours"]]

        # Date filtering.
        engagements_df = date_filter(
            start=start_date, end=end_date, data_frame=engagements_df, date_column="activity_date"
        )

        # Date aggregation.
        engagements_df = granularity_aggregation(
            level=granularity,
            group=["activity_date", "enroll_type"],
            date="activity_date",
            data_frame=engagements_df,
            aggregation_type="sum",
        )

        # Calculating metric.
        engagements = calculation_aggregation(calc=calculation, aggregation_type="sum", data_frame=engagements_df)
        return engagements

    def construct_engagements_over_time(self, engagements_df, start_date, end_date, granularity, calculation):
        """
        Construct engagements over time data.

        Arguments:
            engagements_df {DataFrame} -- DataFrame of engagements
            start_date {datetime} -- Engagement start date in the format 'YYYY-MM-DD'
            end_date {datetime} -- Engagement end date in the format 'YYYY-MM-DD'
            granularity {str} -- Granularity of the data. One of Granularity choices
            calculation {str} -- Calculation of the data. One of Calculation choices
        """
        engagements = self.engagements_over_time_common(engagements_df, start_date, end_date, granularity, calculation)
        # convert dataframe to a list of records
        return engagements.to_dict(orient='records')

    def construct_engagements_over_time_csv(self, engagements_df, start_date, end_date, granularity, calculation):
        """
        Construct engagements over time CSV.

        Arguments:
            engagements_df {DataFrame} -- DataFrame of engagements
            start_date {datetime} -- Engagement start date in the format 'YYYY-MM-DD'
            end_date {datetime} -- Engagement end date in the format 'YYYY-MM-DD'
            granularity {str} -- Granularity of the data. One of Granularity choices
            calculation {str} -- Calculation of the data. One of Calculation choices
        """
        engagements = self.engagements_over_time_common(engagements_df, start_date, end_date, granularity, calculation)

        engagements = engagements.pivot(
            index="activity_date", columns="enroll_type", values="sum"
        )

        filename = f"Engagement Timeseries, {start_date} - {end_date} ({granularity} {calculation}).csv"
        return self.construct_csv_response(engagements, filename)

    def top_courses_by_engagements_common(self, engagements_df, start_date, end_date):
        """
        Common method for constructing top courses by engagements data.

        Arguments:
            engagements_df {DataFrame} -- DataFrame of engagements
            start_date {datetime} -- Engagement start date in the format 'YYYY-MM-DD'
            end_date {datetime} -- Engagement end date in the format 'YYYY-MM-DD'
            group_by_columns {list} -- List of columns to group by
            columns {list} -- List of column for the final result
        """
        engagements_df["learning_time_hours"] = engagements_df["learning_time_seconds"] / 60 / 60
        engagements_df["learning_time_hours"] = engagements_df["learning_time_hours"].astype("float")

        # Date filtering.
        engagements = date_filter(
            start=start_date, end=end_date, data_frame=engagements_df, date_column="activity_date"
        )

        courses = list(
            engagements.groupby(["course_key"])
            .learning_time_hours.sum()
            .sort_values(ascending=False)[:10]
            .index
        )

        engagements = (
            engagements_df[engagements_df.course_key.isin(courses)]
            .groupby(["course_key", "course_title", "enroll_type"])
            .learning_time_hours.sum()
            .reset_index()
        )

        engagements.columns = ["course_key", "course_title", "enroll_type", "count"]

        return engagements

    def construct_top_courses_by_engagements(self, engagements_df, start_date, end_date):
        """
        Construct top courses by engagements data.

        Arguments:
            engagements_df {DataFrame} -- DataFrame of engagements
            start_date {datetime} -- Engagement start date in the format 'YYYY-MM-DD'
            end_date {datetime} -- Engagement end date in the format 'YYYY-MM-DD'
        """
        engagements = self.top_courses_by_engagements_common(
            engagements_df,
            start_date,
            end_date
        )

        # convert dataframe to a list of records
        return engagements.to_dict(orient='records')

    def construct_top_courses_by_engagements_csv(self, engagements_df, start_date, end_date):
        """
        Construct top courses by engagements CSV.

        Arguments:
            engagements_df {DataFrame} -- DataFrame of engagements
            start_date {datetime} -- Engagement start date in the format 'YYYY-MM-DD'
            end_date {datetime} -- Engagement end date in the format 'YYYY-MM-DD'
        """
        engagements = self.top_courses_by_engagements_common(
            engagements_df,
            start_date,
            end_date
        )

        engagements = engagements.pivot(
            index=["course_key", "course_title"], columns="enroll_type", values="count"
        )

        filename = f"Top 10 Courses by Learning Hours, {start_date} - {end_date}.csv"
        return self.construct_csv_response(engagements, filename)

    def top_subjects_by_engagements_common(self, engagements_df, start_date, end_date):
        """
        Common method for constructing top subjects by engagements data.

        Arguments:
            engagements_df {DataFrame} -- DataFrame of engagements
            start_date {datetime} -- Engagement start date in the format 'YYYY-MM-DD'
            end_date {datetime} -- Engagement end date in the format 'YYYY-MM-DD'
        """
        engagements_df["learning_time_hours"] = engagements_df["learning_time_seconds"] / 60 / 60
        engagements_df["learning_time_hours"] = engagements_df["learning_time_hours"].astype("float")

        # Date filtering.
        engagements = date_filter(
            start=start_date, end=end_date, data_frame=engagements_df, date_column="activity_date"
        )

        subjects = list(
            engagements.groupby(["course_subject"])
            .learning_time_hours.sum()
            .sort_values(ascending=False)[:10]
            .index
        )

        engagements = (
            engagements[engagements.course_subject.isin(subjects)]
            .groupby(["course_subject", "enroll_type"])
            .learning_time_hours.sum()
            .reset_index()
        )
        engagements.columns = ["course_subject", "enroll_type", "count"]

        return engagements

    def construct_top_subjects_by_engagements(self, engagements_df, start_date, end_date):
        """
        Construct top subjects by engagements data.

        Arguments:
            engagements_df {DataFrame} -- DataFrame of engagements
            start_date {datetime} -- Engagement start date in the format 'YYYY-MM-DD'
            end_date {datetime} -- Engagement end date in the format 'YYYY-MM-DD'
        """
        engagements = self.top_subjects_by_engagements_common(engagements_df, start_date, end_date)
        # convert dataframe to a list of records
        return engagements.to_dict(orient='records')

    def construct_top_subjects_by_engagements_csv(self, engagements_df, start_date, end_date):
        """
        Construct top subjects by engagements CSV.

        Arguments:
            engagements_df {DataFrame} -- DataFrame of engagements
            start_date {datetime} -- Engagement start date in the format 'YYYY-MM-DD'
            end_date {datetime} -- Engagement end date in the format 'YYYY-MM-DD'
        """
        engagements = self.top_subjects_by_engagements_common(engagements_df, start_date, end_date)
        engagements = engagements.pivot(index="course_subject", columns="enroll_type", values="count")
        filename = f"Top 10 Subjects by Learning Hours, {start_date} - {end_date}.csv"
        return self.construct_csv_response(engagements, filename)

    def construct_csv_response(self, engagements, filename):
        """
        Construct CSV response.

        Arguments:
            engagements {DataFrame} -- DataFrame of engagements
            filename {str} -- Filename for the CSV
        """
        response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = f'attachment; filename="{filename}"'
        response['Access-Control-Expose-Headers'] = 'Content-Disposition'
        engagements.to_csv(path_or_buf=response)

        return response



"""Unittests for analytics_enrollments.py"""

from datetime import datetime

import ddt
from mock import patch
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITransactionTestCase

from enterprise_data.admin_analytics.constants import ResponseType
from enterprise_data.tests.admin_analytics.mock_analytics_data import (
    ENROLLMENTS,
    LEADERBOARD_RESPONSE,
    engagements_dataframe,
    enrollments_dataframe,
    leaderboard_csv_content,
)
from enterprise_data.tests.mixins import JWTTestMixin
from enterprise_data.tests.test_utils import UserFactory
from enterprise_data_roles.constants import ENTERPRISE_DATA_ADMIN_ROLE
from enterprise_data_roles.models import EnterpriseDataFeatureRole, EnterpriseDataRoleAssignment


@ddt.ddt
class TestLeaderboardAPI(JWTTestMixin, APITransactionTestCase):
    """Tests for AdvanceAnalyticsLeaderboardView."""

    def setUp(self):
        """
        Setup method.
        """
        super().setUp()
        self.user = UserFactory(is_staff=True)
        role, __ = EnterpriseDataFeatureRole.objects.get_or_create(
            name=ENTERPRISE_DATA_ADMIN_ROLE
        )
        self.role_assignment = EnterpriseDataRoleAssignment.objects.create(
            role=role, user=self.user
        )
        self.client.force_authenticate(user=self.user)

        self.enterprise_uuid = "ee5e6b3a-069a-4947-bb8d-d2dbc323396c"
        self.set_jwt_cookie()

        self.url = reverse(
            "v1:enterprise-admin-analytics-leaderboard",
            kwargs={"enterprise_uuid": self.enterprise_uuid},
        )

        fetch_max_enrollment_datetime_patcher = patch(
            'enterprise_data.api.v1.views.analytics_leaderboard.fetch_enrollments_cache_expiry_timestamp',
            return_value=datetime.now()
        )

        fetch_max_enrollment_datetime_patcher.start()
        self.addCleanup(fetch_max_enrollment_datetime_patcher.stop)

        fetch_max_engagement_datetime_patcher = patch(
            'enterprise_data.api.v1.views.analytics_leaderboard.fetch_engagements_cache_expiry_timestamp',
            return_value=datetime.now()
        )

        fetch_max_engagement_datetime_patcher.start()
        self.addCleanup(fetch_max_engagement_datetime_patcher.stop)

    def verify_enrollment_data(self, results, results_count):
        """Verify the received enrollment data."""
        attrs = [
            "email",
            "course_title",
            "course_subject",
            "enroll_type",
            "enterprise_enrollment_date",
        ]

        assert len(results) == results_count

        filtered_data = []
        for enrollment in ENROLLMENTS:
            for result in results:
                if enrollment["email"] == result["email"]:
                    filtered_data.append({attr: enrollment[attr] for attr in attrs})
                    break

        received_data = sorted(results, key=lambda x: x["email"])
        expected_data = sorted(filtered_data, key=lambda x: x["email"])
        assert received_data == expected_data

    @patch(
        "enterprise_data.api.v1.views.analytics_leaderboard.fetch_and_cache_enrollments_data"
    )
    @patch(
        "enterprise_data.api.v1.views.analytics_leaderboard.fetch_and_cache_engagements_data"
    )
    def test_get(self, mock_fetch_and_cache_engagements_data, mock_fetch_and_cache_enrollments_data):
        """
        Test the GET method for the AdvanceAnalyticsLeaderboardView works.
        """
        mock_fetch_and_cache_enrollments_data.return_value = enrollments_dataframe()
        mock_fetch_and_cache_engagements_data.return_value = engagements_dataframe()

        response = self.client.get(self.url, {"page_size": 2})
        assert response.status_code == status.HTTP_200_OK
        assert response["Content-Type"] == "application/json"
        data = response.json()
        assert data["next"] == f'http://testserver{self.url}?page=2&page_size=2'
        assert data["previous"] is None
        assert data["current_page"] == 1
        assert data["num_pages"] == 6
        assert data["count"] == 12
        assert data["results"] == [
            {
                "email": "paul77@example.org",
                "daily_sessions": 1,
                "learning_time_seconds": 15753,
                "learning_time_hours": 4.4,
                "average_session_length": 4.4,
                "course_completions": None,
            },
            {
                "email": "seth57@example.org",
                "daily_sessions": 1,
                "learning_time_seconds": 9898,
                "learning_time_hours": 2.7,
                "average_session_length": 2.7,
                "course_completions": None,
            },
        ]

        # fetch all records
        response = self.client.get(self.url, {"page_size": 20})
        assert response.status_code == status.HTTP_200_OK
        data = response.json()
        assert data["next"] is None
        assert data["previous"] is None
        assert data["current_page"] == 1
        assert data["num_pages"] == 1
        assert data["count"] == 12
        assert data["results"] == LEADERBOARD_RESPONSE

    @patch(
        "enterprise_data.api.v1.views.analytics_leaderboard.fetch_and_cache_enrollments_data"
    )
    @patch(
        "enterprise_data.api.v1.views.analytics_leaderboard.fetch_and_cache_engagements_data"
    )
    def test_get_csv(self, mock_fetch_and_cache_engagements_data, mock_fetch_and_cache_enrollments_data):
        """
        Test the GET method for the AdvanceAnalyticsIndividualEnrollmentsView return correct CSV data.
        """
        mock_fetch_and_cache_enrollments_data.return_value = enrollments_dataframe()
        mock_fetch_and_cache_engagements_data.return_value = engagements_dataframe()

        response = self.client.get(self.url, {"response_type": ResponseType.CSV.value})
        assert response.status_code == status.HTTP_200_OK

        # verify the response headers
        assert response["Content-Type"] == "text/csv"

        # verify the response content
        content = b"".join(response.streaming_content)
        assert content == leaderboard_csv_content()



"""Views for enterprise admin completions analytics."""
import datetime
from datetime import datetime, timedelta
from logging import getLogger

from edx_rbac.decorators import permission_required
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from rest_framework.response import Response
from rest_framework.status import HTTP_200_OK
from rest_framework.views import APIView

from django.http import HttpResponse

from enterprise_data.admin_analytics.completions_utils import (
    get_completions_over_time,
    get_csv_data_for_completions_over_time,
    get_csv_data_for_top_courses_by_completions,
    get_csv_data_for_top_subjects_by_completions,
    get_top_courses_by_completions,
    get_top_subjects_by_completions,
)
from enterprise_data.admin_analytics.constants import Calculation, Granularity
from enterprise_data.admin_analytics.data_loaders import fetch_max_enrollment_datetime
from enterprise_data.admin_analytics.utils import ChartType, fetch_and_cache_enrollments_data
from enterprise_data.api.v1 import serializers
from enterprise_data.api.v1.paginators import AdvanceAnalyticsPagination
from enterprise_data.utils import date_filter, timer

LOGGER = getLogger(__name__)


class EnterrpiseAdminCompletionsStatsView(APIView):
    """
    API for getting the enterprise admin completions.
    """
    authentication_classes = (JwtAuthentication,)
    http_method_names = ['get']

    @permission_required(
        "can_access_enterprise", fn=lambda request, enterprise_id: enterprise_id
    )
    def get(self, request, enterprise_id):
        """
        HTTP GET endpoint to retrieve the enterprise admin completions
        """
        serializer = serializers.AdminAnalyticsAggregatesQueryParamsSerializer(
            data=request.GET
        )
        serializer.is_valid(raise_exception=True)

        last_updated_at = fetch_max_enrollment_datetime()
        cache_expiry = (
            last_updated_at + timedelta(days=1) if last_updated_at else datetime.now()
        )

        enrollments = fetch_and_cache_enrollments_data(
            enterprise_id, cache_expiry
        ).copy()
        # Use start and end date if provided by the client, if client has not provided then use
        # 1. minimum enrollment date from the data as the start_date
        # 2. today's date as the end_date
        start_date = serializer.data.get(
            "start_date", enrollments.enterprise_enrollment_date.min()
        )
        end_date = serializer.data.get("end_date", datetime.now())

        if serializer.data.get('response_type') == 'csv':
            chart_type = serializer.data.get('chart_type')
            response = HttpResponse(content_type='text/csv')
            csv_data = {}

            if chart_type == ChartType.COMPLETIONS_OVER_TIME.value:
                with timer('completions_over_time_csv_data'):
                    csv_data = get_csv_data_for_completions_over_time(
                        start_date=start_date,
                        end_date=end_date,
                        enrollments=enrollments.copy(),
                        date_agg=serializer.data.get('granularity', Granularity.DAILY.value),
                        calc=serializer.data.get('calculation', Calculation.TOTAL.value),
                    )
            elif chart_type == ChartType.TOP_COURSES_BY_COMPLETIONS.value:
                with timer('top_courses_by_completions_csv_data'):
                    csv_data = get_csv_data_for_top_courses_by_completions(
                        start_date=start_date, end_date=end_date, enrollments=enrollments.copy()
                    )
            elif chart_type == ChartType.TOP_SUBJECTS_BY_COMPLETIONS.value:
                with timer('top_subjects_by_completions_csv_data'):
                    csv_data = get_csv_data_for_top_subjects_by_completions(
                        start_date=start_date, end_date=end_date, enrollments=enrollments.copy()
                    )
            filename = csv_data['filename']
            response['Content-Disposition'] = f'attachment; filename="{filename}"'
            response['Access-Control-Expose-Headers'] = 'Content-Disposition'
            csv_data['data'].to_csv(path_or_buf=response)
            return response

        with timer('completions_all_charts_data'):
            completions_over_time = get_completions_over_time(
                start_date=start_date,
                end_date=end_date,
                dff=enrollments.copy(),
                date_agg=serializer.data.get('granularity', Granularity.DAILY.value),
                calc=serializer.data.get('calculation', Calculation.TOTAL.value),
            )
            top_courses_by_completions = get_top_courses_by_completions(
                start_date=start_date, end_date=end_date, dff=enrollments.copy()
            )
            top_subjects_by_completions = get_top_subjects_by_completions(
                start_date=start_date, end_date=end_date, dff=enrollments.copy()
            )

        return Response(
            data={
                'completions_over_time': completions_over_time.to_dict(
                    orient="records"
                ),
                'top_courses_by_completions': top_courses_by_completions.to_dict(
                    orient="records"
                ),
                'top_subjects_by_completions': top_subjects_by_completions.to_dict(
                    orient="records"
                ),
            },
            status=HTTP_200_OK,
        )


class EnterrpiseAdminCompletionsView(APIView):
    """
    API for getting the enterprise admin completions.
    """
    authentication_classes = (JwtAuthentication,)
    http_method_names = ['get']
    pagination_class = AdvanceAnalyticsPagination

    @permission_required(
        "can_access_enterprise", fn=lambda request, enterprise_id: enterprise_id
    )
    def get(self, request, enterprise_id):
        """
        HTTP GET endpoint to retrieve the enterprise admin completions
        """
        serializer = serializers.AdminAnalyticsAggregatesQueryParamsSerializer(
            data=request.GET
        )
        serializer.is_valid(raise_exception=True)

        last_updated_at = fetch_max_enrollment_datetime()
        cache_expiry = (
            last_updated_at + timedelta(days=1) if last_updated_at else datetime.now()
        )

        enrollments = fetch_and_cache_enrollments_data(
            enterprise_id, cache_expiry
        ).copy()
        # Use start and end date if provided by the client, if client has not provided then use
        # 1. minimum enrollment date from the data as the start_date
        # 2. today's date as the end_date
        start_date = serializer.data.get(
            'start_date', enrollments.enterprise_enrollment_date.min()
        )
        end_date = serializer.data.get('end_date', datetime.now())

        LOGGER.info(
            "Completions data requested for enterprise [%s] from [%s] to [%s]",
            enterprise_id,
            start_date,
            end_date,
        )

        dff = enrollments[enrollments['has_passed'] == 1]

        # Date filtering
        dff = date_filter(start=start_date, end=end_date, data_frame=dff, date_column='passed_date')

        dff = dff[['email', 'course_title', 'course_subject', 'passed_date']]
        dff['passed_date'] = dff['passed_date'].dt.date
        dff = dff.sort_values(by="passed_date", ascending=False).reset_index(drop=True)

        LOGGER.info(
            "Completions data prepared for enterprise [%s] from [%s] to [%s]",
            enterprise_id,
            start_date,
            end_date,
        )

        if serializer.data.get('response_type') == 'csv':
            response = HttpResponse(content_type='text/csv')
            filename = f"Individual Completions, {start_date} - {end_date}.csv"
            response['Content-Disposition'] = f'attachment; filename="{filename}"'
            response['Access-Control-Expose-Headers'] = 'Content-Disposition'
            dff.to_csv(path_or_buf=response, index=False)
            return response

        paginator = self.pagination_class()
        page = paginator.paginate_queryset(dff, request)
        serialized_data = page.data.to_dict(orient='records')
        response = paginator.get_paginated_response(serialized_data)

        return response



"""Custom paginator for the Advance Analytics API."""

import math
from dataclasses import dataclass
from typing import Any

from rest_framework.exceptions import NotFound
from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response


@dataclass
class Page:
    """
    A class representing a single page of paginated data.

    Attributes:
        data (Any): The data contained in the current page.
        count (int): The total number of items across all pages.
        num_pages (int): The total number of pages.
        current_page (int): The current page number.
    """
    data: Any
    count: int
    num_pages: int
    current_page: int

    def has_next(self):
        """
        Check if there is a next page.

        Returns:
            bool: True if there is a next page, False otherwise.
        """
        return self.current_page < self.num_pages

    def has_previous(self):
        """
        Check if there is a previous page.

        Returns:
            bool: True if there is a previous page, False otherwise.
        """
        return self.current_page > 1

    def next_page_number(self):
        """
        Get the next page number.

        Returns:
            int: The next page number.
        """
        return self.current_page + 1

    def previous_page_number(self):
        """
        Get the previous page number.

        Returns:
            int: The previous page number.
        """
        return self.current_page - 1


class AdvanceAnalyticsPagination(PageNumberPagination):
    """
    Custom pagination class for advanced analytics.

    Attributes:
        page_size_query_param (str): The query parameter for the page size.
        page_size (int): The default page size.
        max_page_size (int): The maximum allowed page size.
    """
    page_size_query_param = "page_size"
    page_size = 50
    max_page_size = 100

    def paginate_queryset(self, queryset, request, view=None):
        """
        Paginate a given dataframe based on the request parameters.

        Args:
            queryset (pd.DataFrame): The dataframe to paginate.
            request (Request): The request object containing query parameters.
            view (View, optional): The view that is calling the paginator.

        Returns:
            Page: A Page object. `data` attribute of the object will contain the paginated data.
        """
        dataframe = queryset

        self.request = request  # pylint: disable=attribute-defined-outside-init
        page_size = self.get_page_size(request)
        if not page_size:
            return None

        total_rows = dataframe.shape[0]
        num_pages = math.ceil(total_rows / page_size)

        page_number = int(request.query_params.get(self.page_query_param) or 1)
        if page_number <= 0 or page_number > num_pages:
            raise NotFound('Invalid page.')

        start_index = (page_number - 1) * page_size
        end_index = min(start_index + page_size, total_rows)
        data_frame_page = dataframe.iloc[start_index:end_index]

        # pylint: disable=attribute-defined-outside-init
        self.page = Page(data_frame_page, total_rows, num_pages, page_number)

        return self.page

    def get_paginated_response(self, data):
        return Response({
            'next': self.get_next_link(),
            'previous': self.get_previous_link(),
            'count': self.page.count,
            'num_pages': self.page.num_pages,
            'current_page': self.page.current_page,
            'results': data
        })



"""
Module containing queries for the fact_enrollment_engagement_day_admin_dash table.
"""


class FactEngagementAdminDashQueries:
    """
    Queries related to the fact_enrollment_engagement_day_admin_dash table.
    """
    @staticmethod
    def get_learning_hours_and_daily_sessions_query():
        """
        Get the query to fetch the learning hours and daily sessions.
        """
        return """
            SELECT
                ROUND(SUM(learning_time_seconds) / 60 / 60, 1) as hours, SUM(is_engaged) as sessions
            FROM fact_enrollment_engagement_day_admin_dash
            WHERE enterprise_customer_uuid=%(enterprise_customer_uuid)s AND
                activity_date BETWEEN %(start_date)s AND %(end_date)s;
        """



"""This module contains utility functions for completions analytics."""
from enterprise_data.utils import date_filter


def date_aggregation(level, group, date, df, type_="count"):
    """Perform date aggregation on a DataFrame.

    This function aggregates data based on the specified level of aggregation (e.g., daily, weekly, monthly, quarterly)
    and returns the aggregated data.

    Args:
        level (str): The level of aggregation. Possible values are "Daily", "Weekly", "Monthly", and "Quarterly".
        group (list): A list of column names to group the data by.
        date (str): The name of the date column in the DataFrame.
        df (pandas.DataFrame): The DataFrame containing the data to be aggregated.
        type_ (str, optional): The type of aggregation to perform. Possible values
            are "count" and "sum". Defaults to "count".

    Returns:
        pandas.DataFrame: The aggregated data.

    """
    if type_ == "count":
        if level == "Daily":
            df = df.groupby(group).size().reset_index()
            group.append("count")
            df.columns = group
        elif level == "Weekly":
            df[date] = df[date].dt.to_period("W").dt.start_time
            df = df.groupby(group).size().reset_index()
            group.append("count")
            df.columns = group
        elif level == "Monthly":
            df[date] = df[date].dt.to_period("M").dt.start_time
            df = df.groupby(group).size().reset_index()
            group.append("count")
            df.columns = group
        elif level == "Quarterly":
            df[date] = df[date].dt.to_period("Q").dt.start_time
            df = df.groupby(group).size().reset_index()
            group.append("count")
            df.columns = group
    elif type_ == "sum":
        if level == "Daily":
            df = df.groupby(group).sum().reset_index()
            group.append("sum")
            df.columns = group
        elif level == "Weekly":
            df[date] = df[date].dt.to_period("W").dt.start_time
            df = df.groupby(group).sum().reset_index()
            group.append("sum")
            df.columns = group
        elif level == "Monthly":
            df[date] = df[date].dt.to_period("M").dt.start_time
            df = df.groupby(group).sum().reset_index()
            group.append("sum")
            df.columns = group
        elif level == "Quarterly":
            df[date] = df[date].dt.to_period("Q").dt.start_time
            df = df.groupby(group).sum().reset_index()
            group.append("sum")
            df.columns = group

    return df


def calculation(calc, df, type_="count"):
    """Perform a calculation on the given DataFrame based on the specified calculation type.

    Args:
        calc (str): The calculation type. Possible values are "Total", "Running Total",
                    "Moving Average (3 Period)", and "Moving Average (7 Period)".
        df (pandas.DataFrame): The filtered enrollments data.
        type_ (str, optional): The type of calculation to perform. Default is "count".

    Returns:
        pandas.DataFrame: The aggregated data after performing the calculation.
    """
    if type_ == "count":
        if calc == "Total":
            pass
        elif calc == "Running Total":
            df["count"] = df.groupby("enroll_type")["count"].cumsum()
        elif calc == "Moving Average (3 Period)":
            df["count"] = (
                df.groupby("enroll_type")["count"]
                .rolling(3)
                .mean()
                .droplevel(level=[0])
            )
        elif calc == "Moving Average (7 Period)":
            df["count"] = (
                df.groupby("enroll_type")["count"]
                .rolling(7)
                .mean()
                .droplevel(level=[0])
            )
    elif type_ == "sum":
        if calc == "Total":
            pass
        elif calc == "Running Total":
            df["sum"] = df.groupby("enroll_type")["sum"].cumsum()
        elif calc == "Moving Average (3 Period)":
            df["sum"] = (
                df.groupby("enroll_type")["sum"].rolling(3).mean().droplevel(level=[0])
            )
        elif calc == "Moving Average (7 Period)":
            df["sum"] = (
                df.groupby("enroll_type")["sum"].rolling(7).mean().droplevel(level=[0])
            )

    return df


def get_completions_over_time(start_date, end_date, dff, date_agg, calc):
    """Get agreggated data for completions over time graph.

    Args:
        start_date (datetime): The start date for the date filter.
        end_date (datetime): The end date for the date filter.
        dff (pandas.DataFrame): enrollments data
        date_agg (str): It denotes the granularity of the aggregated date which can be Daily, Weekly, Monthly, Quarterly
        calc (str): Calculations denoiated the period for the running averages. It can be Total, Running Total, Moving
            Average (3 Period), Moving Average (7 Period)
    """

    dff = dff[dff["has_passed"] == 1]

    # Date filtering.
    dff = date_filter(start=start_date, end=end_date, data_frame=dff, date_column="passed_date")

    # Date aggregation.
    dff = date_aggregation(
        level=date_agg, group=["passed_date", "enroll_type"], date="passed_date", df=dff
    )

    # Calculating metric.
    dff = calculation(calc=calc, df=dff)

    return dff


def get_top_courses_by_completions(start_date, end_date, dff):
    """Get top 10 courses by completions.

    Args:
        start_date (datetime): The start date for the date filter.
        end_date (datetime): The end date for the date filter.
        dff (pandas.DataFrame): Enrollments data
    """

    dff = dff[dff["has_passed"] == 1]

    # Date filtering.
    dff = date_filter(start=start_date, end=end_date, data_frame=dff, date_column="passed_date")

    courses = list(
        dff.groupby(["course_key"]).size().sort_values(ascending=False)[:10].index
    )

    dff = (
        dff[dff.course_key.isin(courses)]
        .groupby(["course_key", "course_title", "enroll_type"])
        .size()
        .reset_index()
    )
    dff.columns = ["course_key", "course_title", "enroll_type", "count"]

    return dff


def get_top_subjects_by_completions(start_date, end_date, dff):
    """Get top 10 subjects by completions.

    Args:
        start_date (datetime): The start date for the date filter.
        end_date (datetime): The end date for the date filter.
        dff (pandas.DataFrame): Enrollments data
    """

    dff = dff[dff["has_passed"] == 1]

    # Date filtering.
    dff = date_filter(start=start_date, end=end_date, data_frame=dff, date_column="passed_date")

    subjects = list(
        dff.groupby(["course_subject"]).size().sort_values(ascending=False)[:10].index
    )

    dff = (
        dff[dff.course_subject.isin(subjects)]
        .groupby(["course_subject", "enroll_type"])
        .size()
        .reset_index()
    )
    dff.columns = ["course_subject", "enroll_type", "count"]

    return dff


def get_csv_data_for_completions_over_time(
    start_date, end_date, enrollments, date_agg, calc
):
    """Get csv data for completions over time graph.

    Args:
        start_date (datetime): The start date for the date filter.
        end_date (datetime): The end date for the date filter.
        enrollments (pandas.DataFrame): Filtered enrollments data
        date_agg (str): it denotes the granularity of the aggregated date which can be Daily, Weekly, Monthly, Quarterly
        calc (str): calculations denoiated the period for the running averages. It can be Total, Running Total, Moving
            Average (3 Period), Moving Average (7 Period)

    Returns:
        dict: csv data
    """

    dff = get_completions_over_time(start_date, end_date, enrollments, date_agg, calc)
    dff = dff.pivot(index="passed_date", columns="enroll_type", values="count")
    filename = (
        f"Completions Timeseries, {start_date} - {end_date} ({date_agg} {calc}).csv"
    )
    return {"filename": filename, "data": dff}


def get_csv_data_for_top_courses_by_completions(start_date, end_date, enrollments):
    """Get csv data for top 10 courses by completions.

    Args:
        start_date (datetime): The start date for the date filter.
        end_date (datetime): The end date for the date filter.
        enrollments (pandas.DataFrame): Filtered enrollments data

    Returns:
        dict: csv data
    """

    dff = get_top_courses_by_completions(start_date, end_date, enrollments)
    dff = dff.pivot(
        index=["course_key", "course_title"], columns="enroll_type", values="count"
    )
    filename = f"Top 10 Courses by Completions, {start_date} - {end_date}.csv"
    return {"filename": filename, "data": dff}


def get_csv_data_for_top_subjects_by_completions(start_date, end_date, enrollments):
    """Get csv data for top 10 subjects by completions.

    Args:
        start_date (datetime): The start date for the date filter.
        end_date (datetime): The end date for the date filter.
        enrollments (pandas.DataFrame): Filtered enrollments data

    Returns:
        dict: csv data
    """

    dff = get_top_subjects_by_completions(start_date, end_date, enrollments)
    dff = dff.pivot(index="course_subject", columns="enroll_type", values="count")
    filename = f"Top 10 Subjects by Completions, {start_date} - {end_date}.csv"
    return {"filename": filename, "data": dff}



"""
Module containing queries for the fact_enrollment_admin_dash table.
"""


class FactEnrollmentAdminDashQueries:
    """
    Queries related to the fact_enrollment_admin_dash table.
    """
    @staticmethod
    def get_enrollment_count_query():
        """
        Get the query to fetch the total number of enrollments for an enterprise customer.
        """
        return """
            SELECT count(*)
            FROM fact_enrollment_admin_dash
            WHERE enterprise_customer_uuid=%(enterprise_customer_uuid)s AND
                enterprise_enrollment_date BETWEEN %(start_date)s AND %(end_date)s;
        """

    @staticmethod
    def get_all_enrollments_query():
        """
        Get the query to fetch all enrollments.
        """
        return """
            SELECT email, course_title, course_subject, enroll_type, enterprise_enrollment_date
            FROM fact_enrollment_admin_dash
            WHERE enterprise_customer_uuid=%(enterprise_customer_uuid)s AND
                enterprise_enrollment_date BETWEEN %(start_date)s AND %(end_date)s
            ORDER BY ENTERPRISE_ENROLLMENT_DATE DESC LIMIT %(limit)s OFFSET %(offset)s
        """

    @staticmethod
    def get_enrollment_date_range_query():
        """
        Get the query to fetch the enrollment date range.
        """
        return """
            SELECT
                MIN(enterprise_enrollment_date) AS min_enrollment_date,
                MAX(enterprise_enrollment_date) AS max_enrollment_date
            FROM fact_enrollment_admin_dash
            WHERE enterprise_customer_uuid=%(enterprise_customer_uuid)s;
        """

    @staticmethod
    def get_enrollment_and_course_count_query():
        """
        Get the query to fetch the enrollment and course count.
        """
        return """
            SELECT
                count(*) as enrolls, count(DISTINCT course_key) as courses
            FROM fact_enrollment_admin_dash
            WHERE enterprise_customer_uuid=%(enterprise_customer_uuid)s AND
                enterprise_enrollment_date BETWEEN %(start_date)s AND %(end_date)s;
        """

    @staticmethod
    def get_completion_count_query():
        """
        Get the query to fetch the completion count.
        """
        return """
            SELECT
                SUM(has_passed) as completions
            FROM fact_enrollment_admin_dash
            WHERE enterprise_customer_uuid=%(enterprise_customer_uuid)s AND
                passed_date BETWEEN %(start_date)s AND %(end_date)s;
        """

    @staticmethod
    def get_learning_hours_and_daily_sessions_query():
        """
        Get the query to fetch the learning hours and daily sessions.
        """
        return """
            SELECT
                ROUND(SUM(learning_time_seconds) / 60 / 60, 1) as hours, SUM(is_engaged) as sessions
            FROM fact_engagement_admin_dash
            WHERE enterprise_customer_uuid=%(enterprise_customer_uuid)s AND
                activity_date BETWEEN %(start_date)s AND %(end_date)s;
        """

    @staticmethod
    def get_top_courses_enrollments_query(record_count=20):
        """
        Get the query to fetch the enrollment count by courses.

        Query will fetch the top N courses by enrollment count. Where N is the value of record_count.

        Arguments:
            record_count (int): Number of records to fetch.
        """
        # Some local environments raise error when course_title is added in SELECT without GROUP BY.
        # If you face this issue, you can remove course_title from SELECT.
        return f"""
            SELECT course_key, course_title , enroll_type, count(course_key) as enrollment_count
            FROM fact_enrollment_admin_dash
            WHERE enterprise_customer_uuid=%(enterprise_customer_uuid)s AND
                enterprise_enrollment_date BETWEEN %(start_date)s AND %(end_date)s
            GROUP BY course_key, enroll_type ORDER BY enrollment_count DESC LIMIT {record_count};
        """

    @staticmethod
    def get_top_subjects_by_enrollments_query(record_count=20):
        """
        Get the query to fetch the enrollment count by subjects.

        Query will fetch the top N subjects by enrollment count. Where N is the value of record_count.

        Arguments:
            record_count (int): Number of records to fetch.
        """
        return f"""
            SELECT course_subject, enroll_type, count(course_subject) enrollment_count
            FROM fact_enrollment_admin_dash
            WHERE enterprise_customer_uuid=%(enterprise_customer_uuid)s AND
                enterprise_enrollment_date BETWEEN %(start_date)s AND %(end_date)s
            GROUP BY course_subject, enroll_type ORDER BY enrollment_count DESC LIMIT {record_count};
        """

    @staticmethod
    def get_enrolment_time_series_data_query():
        """
        Get the query to fetch the enrollment time series data with daily aggregation.
        """
        return """
            SELECT enterprise_enrollment_date, enroll_type, COUNT(*) as enrollment_count
            FROM fact_enrollment_admin_dash
            WHERE enterprise_customer_uuid=%(enterprise_customer_uuid)s AND
                enterprise_enrollment_date BETWEEN %(start_date)s AND %(end_date)s
            GROUP BY enterprise_enrollment_date, enroll_type
            ORDER BY enterprise_enrollment_date;
        """



"""
Test cases for enterprise_admin views
"""
from datetime import datetime
from unittest import mock
from uuid import uuid4

import ddt
from mock import patch
from pytest import mark
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITransactionTestCase

from enterprise_data.admin_analytics.database.queries import (
    FactEngagementAdminDashQueries,
    FactEnrollmentAdminDashQueries,
)
from enterprise_data.tests.mixins import JWTTestMixin
from enterprise_data.tests.test_utils import (
    UserFactory,
    get_dummy_enrollments_data,
    get_dummy_enterprise_api_data,
    get_dummy_skills_data,
)
from enterprise_data_roles.constants import ENTERPRISE_DATA_ADMIN_ROLE
from enterprise_data_roles.models import EnterpriseDataFeatureRole, EnterpriseDataRoleAssignment


@ddt.ddt
@mark.django_db
class TestEnterpriseAdminAnalyticsAggregatesView(JWTTestMixin, APITransactionTestCase):
    """
    Tests for EnterpriseAdminAnalyticsAggregatesView.
    """

    def setUp(self):
        """
        Setup method.
        """
        super().setUp()
        self.user = UserFactory(is_staff=True)
        role, __ = EnterpriseDataFeatureRole.objects.get_or_create(name=ENTERPRISE_DATA_ADMIN_ROLE)
        self.role_assignment = EnterpriseDataRoleAssignment.objects.create(
            role=role,
            user=self.user
        )
        self.client.force_authenticate(user=self.user)

        mocked_get_enterprise_customer = mock.patch(
            'enterprise_data.filters.EnterpriseApiClient.get_enterprise_customer',
            return_value=get_dummy_enterprise_api_data()
        )

        self.mocked_get_enterprise_customer = mocked_get_enterprise_customer.start()
        self.addCleanup(mocked_get_enterprise_customer.stop)
        self.enterprise_id = 'ee5e6b3a-069a-4947-bb8d-d2dbc323396c'
        self.set_jwt_cookie()
        self.enrollment_queries = FactEnrollmentAdminDashQueries()
        self.engagement_queries = FactEngagementAdminDashQueries()

    def _mock_run_query(self, query, *args, **kwargs):
        """
        mock implementation of run_query.
        """
        mock_responses = {
            self.enrollment_queries.get_enrollment_date_range_query(): [[
                datetime.strptime('2021-01-01', "%Y-%m-%d"),
                datetime.strptime('2021-12-31', "%Y-%m-%d"),
            ]],
            self.enrollment_queries.get_enrollment_and_course_count_query(): [[
                100, 10
            ]],
            self.enrollment_queries.get_completion_count_query(): [[
                50
            ]],
            self.engagement_queries.get_learning_hours_and_daily_sessions_query(): [[
                100, 10
            ]],
            'SELECT MAX(created) FROM enterprise_learner_enrollment': [[datetime.strptime('2021-01-01', "%Y-%m-%d")]]
        }
        return mock_responses[query]

    def test_get_admin_analytics_aggregates(self):
        """
        Test to get admin analytics aggregates.
        """
        url = reverse('v1:enterprise-admin-analytics-aggregates', kwargs={'enterprise_id': self.enterprise_id})
        with patch('enterprise_data.admin_analytics.data_loaders.run_query', side_effect=self._mock_run_query):
            with patch(
                    'enterprise_data.admin_analytics.database.tables.fact_engagement_admin_dash.run_query',
                    side_effect=self._mock_run_query
            ):
                with patch(
                        'enterprise_data.admin_analytics.database.tables.fact_enrollment_admin_dash.run_query',
                        side_effect=self._mock_run_query
                ):
                    response = self.client.get(url)
                    assert response.status_code == status.HTTP_200_OK
                    assert 'enrolls' in response.json()
                    assert 'courses' in response.json()
                    assert 'completions' in response.json()
                    assert 'hours' in response.json()
                    assert 'sessions' in response.json()
                    assert 'last_updated_at' in response.json()
                    assert 'min_enrollment_date' in response.json()
                    assert 'max_enrollment_date' in response.json()


@ddt.ddt
@mark.django_db
class TestEnterpriseAdminAnalyticsSkillsView(JWTTestMixin, APITransactionTestCase):
    """
    Tests for EnterpriseAdminAnalyticsSkillsView.
    """

    def setUp(self):
        """
        Setup method.
        """
        super().setUp()
        self.user = UserFactory()
        self.enterprise_id = uuid4()
        self.set_jwt_cookie(context=f'{self.enterprise_id}')

    def _mock_run_query(self, query):
        """
        mock implementation of run_query.
        """
        if 'skills_daily_rollup_admin_dash' in query:
            return [
                list(item.values()) for item in get_dummy_skills_data(self.enterprise_id)
            ]
        elif 'fact_enrollment_admin_dash' in query:
            return [
                list(item.values()) for item in get_dummy_enrollments_data(self.enterprise_id, 15)
            ]
        else:
            return [
                list(item.values()) for item in get_dummy_skills_data(self.enterprise_id)
            ]

    def test_get_admin_analytics_skills(self):
        """
        Test to get admin analytics skills.
        """
        params = {'start_date': '2021-01-01', 'end_date': '2025-12-31'}
        url = reverse('v1:enterprise-admin-analytics-skills', kwargs={'enterprise_id': self.enterprise_id})
        with patch('enterprise_data.admin_analytics.data_loaders.run_query', side_effect=self._mock_run_query):
            response = self.client.get(url, params)
            assert response.status_code == status.HTTP_200_OK
            data = response.json()
            assert 'top_skills' in data
            assert 'top_skills_by_enrollments' in data
            assert 'top_skills_by_completions' in data
            assert len(data['top_skills']) == 10
            assert len(data['top_skills_by_enrollments']) == 10
            assert len(data['top_skills_by_completions']) == 10



"""Unittests for analytics_enrollments.py"""

from datetime import datetime

import ddt
from mock import patch
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITransactionTestCase

from enterprise_data.admin_analytics.constants import EngagementChart, ResponseType
from enterprise_data.api.v1.serializers import AdvanceAnalyticsEngagementStatsSerializer as EngagementSerializer
from enterprise_data.tests.admin_analytics.mock_analytics_data import (
    ENGAGEMENT_STATS_CSVS,
    ENGAGEMENTS,
    engagements_csv_content,
    engagements_dataframe,
    enrollments_dataframe,
)
from enterprise_data.tests.mixins import JWTTestMixin
from enterprise_data.tests.test_utils import UserFactory
from enterprise_data_roles.constants import ENTERPRISE_DATA_ADMIN_ROLE
from enterprise_data_roles.models import EnterpriseDataFeatureRole, EnterpriseDataRoleAssignment

INVALID_CALCULATION_ERROR = (
    f"Calculation must be one of {EngagementSerializer.CALCULATION_CHOICES}"
)
INVALID_GRANULARITY_ERROR = (
    f"Granularity must be one of {EngagementSerializer.GRANULARITY_CHOICES}"
)
INVALID_CSV_ERROR1 = f"chart_type must be one of {EngagementSerializer.CHART_TYPES}"


@ddt.ddt
class TestIndividualEngagementsAPI(JWTTestMixin, APITransactionTestCase):
    """Tests for AdvanceAnalyticsIndividualEngagementsView."""

    def setUp(self):
        """
        Setup method.
        """
        super().setUp()
        self.user = UserFactory(is_staff=True)
        role, __ = EnterpriseDataFeatureRole.objects.get_or_create(
            name=ENTERPRISE_DATA_ADMIN_ROLE
        )
        self.role_assignment = EnterpriseDataRoleAssignment.objects.create(
            role=role, user=self.user
        )
        self.client.force_authenticate(user=self.user)

        self.enterprise_uuid = "ee5e6b3a-069a-4947-bb8d-d2dbc323396c"
        self.set_jwt_cookie()

        self.url = reverse(
            "v1:enterprise-admin-analytics-engagements",
            kwargs={"enterprise_uuid": self.enterprise_uuid},
        )

        fetch_max_enrollment_datetime_patcher = patch(
            'enterprise_data.admin_analytics.utils.fetch_max_enrollment_datetime',
            return_value=datetime.now()
        )

        fetch_max_enrollment_datetime_patcher.start()
        self.addCleanup(fetch_max_enrollment_datetime_patcher.stop)

    def verify_engagement_data(self, results, results_count):
        """Verify the received engagement data."""
        attrs = [
            "email",
            "course_title",
            "activity_date",
            "course_subject",
        ]

        assert len(results) == results_count

        filtered_data = []
        for engagement in ENGAGEMENTS:
            for result in results:
                if engagement["email"] == result["email"]:
                    data = {attr: engagement[attr] for attr in attrs}
                    data["learning_time_hours"] = round(engagement["learning_time_seconds"] / 3600, 1)
                    filtered_data.append(data)
                    break

        received_data = sorted(results, key=lambda x: x["email"])
        expected_data = sorted(filtered_data, key=lambda x: x["email"])
        assert received_data == expected_data

    @patch(
        "enterprise_data.api.v1.views.analytics_engagements.fetch_and_cache_enrollments_data"
    )
    @patch(
        "enterprise_data.api.v1.views.analytics_engagements.fetch_and_cache_engagements_data"
    )
    def test_get(self, mock_fetch_and_cache_engagements_data, mock_fetch_and_cache_enrollments_data):
        """
        Test the GET method for the AdvanceAnalyticsIndividualEngagementsView works.
        """
        mock_fetch_and_cache_enrollments_data.return_value = enrollments_dataframe()
        mock_fetch_and_cache_engagements_data.return_value = engagements_dataframe()

        response = self.client.get(self.url, {"page_size": 2})
        assert response.status_code == status.HTTP_200_OK
        data = response.json()
        assert data["next"] == f"http://testserver{self.url}?page=2&page_size=2"
        assert data["previous"] is None
        assert data["current_page"] == 1
        assert data["num_pages"] == 5
        assert data["count"] == 9
        self.verify_engagement_data(data["results"], 2)

        response = self.client.get(self.url, {"page_size": 2, "page": 2})
        assert response.status_code == status.HTTP_200_OK
        data = response.json()
        assert data["next"] == f"http://testserver{self.url}?page=3&page_size=2"
        assert data["previous"] == f"http://testserver{self.url}?page_size=2"
        assert data["current_page"] == 2
        assert data["num_pages"] == 5
        assert data["count"] == 9
        self.verify_engagement_data(data["results"], 2)

        response = self.client.get(self.url, {"page_size": 2, "page": 5})
        assert response.status_code == status.HTTP_200_OK
        data = response.json()
        assert data["next"] is None
        assert data["previous"] == f"http://testserver{self.url}?page=4&page_size=2"
        assert data["current_page"] == 5
        assert data["num_pages"] == 5
        assert data["count"] == 9
        self.verify_engagement_data(data["results"], 1)

        response = self.client.get(self.url, {"page_size": 9})
        assert response.status_code == status.HTTP_200_OK
        data = response.json()
        assert data["next"] is None
        assert data["previous"] is None
        assert data["current_page"] == 1
        assert data["num_pages"] == 1
        assert data["count"] == 9
        self.verify_engagement_data(data["results"], 9)

    @patch(
        "enterprise_data.api.v1.views.analytics_engagements.fetch_and_cache_enrollments_data"
    )
    @patch(
        "enterprise_data.api.v1.views.analytics_engagements.fetch_and_cache_engagements_data"
    )
    def test_get_csv(self, mock_fetch_and_cache_engagements_data, mock_fetch_and_cache_enrollments_data):
        """
        Test the GET method for the AdvanceAnalyticsIndividualEngagementsView return correct CSV data.
        """
        mock_fetch_and_cache_enrollments_data.return_value = enrollments_dataframe()
        mock_fetch_and_cache_engagements_data.return_value = engagements_dataframe()
        start_date = enrollments_dataframe().enterprise_enrollment_date.min().strftime('%Y/%m/%d')
        end_date = datetime.now().strftime('%Y/%m/%d')
        response = self.client.get(self.url, {"response_type": ResponseType.CSV.value})
        assert response.status_code == status.HTTP_200_OK

        # verify the response headers
        assert response["Content-Type"] == "text/csv"
        filename = f"""Individual Engagements, {start_date} - {end_date}.csv"""
        assert (
            response["Content-Disposition"] == f'attachment; filename="{filename}"'
        )

        # verify the response content
        content = b"".join(response.streaming_content)
        assert content == engagements_csv_content()

    @ddt.data(
        {
            "params": {"start_date": 1},
            "error": {
                "start_date": [
                    "Date has wrong format. Use one of these formats instead: YYYY-MM-DD."
                ]
            },
        },
        {
            "params": {"end_date": 2},
            "error": {
                "end_date": [
                    "Date has wrong format. Use one of these formats instead: YYYY-MM-DD."
                ]
            },
        },
        {
            "params": {"start_date": "2024-01-01", "end_date": "2023-01-01"},
            "error": {
                "non_field_errors": [
                    "start_date should be less than or equal to end_date."
                ]
            },
        },
        {
            "params": {"calculation": "invalid"},
            "error": {"calculation": [INVALID_CALCULATION_ERROR]},
        },
        {
            "params": {"granularity": "invalid"},
            "error": {"granularity": [INVALID_GRANULARITY_ERROR]},
        },
        {"params": {"chart_type": "invalid"}, "error": {"chart_type": [INVALID_CSV_ERROR1]}},
    )
    @ddt.unpack
    def test_get_invalid_query_params(self, params, error):
        """
        Test the GET method return correct error if any query param value is incorrect.
        """
        response = self.client.get(self.url, params)
        assert response.status_code == status.HTTP_400_BAD_REQUEST
        assert response.json() == error


@ddt.ddt
class TestEngagementStatsAPI(JWTTestMixin, APITransactionTestCase):
    """Tests for AdvanceAnalyticsEngagementStatsView."""

    def setUp(self):
        """
        Setup method.
        """
        super().setUp()
        self.user = UserFactory(is_staff=True)
        role, __ = EnterpriseDataFeatureRole.objects.get_or_create(
            name=ENTERPRISE_DATA_ADMIN_ROLE
        )
        self.role_assignment = EnterpriseDataRoleAssignment.objects.create(
            role=role, user=self.user
        )
        self.client.force_authenticate(user=self.user)

        self.enterprise_uuid = "ee5e6b3a-069a-4947-bb8d-d2dbc323396c"
        self.set_jwt_cookie()

        self.url = reverse(
            "v1:enterprise-admin-analytics-engagements-stats",
            kwargs={"enterprise_uuid": self.enterprise_uuid},
        )

        fetch_max_enrollment_datetime_patcher = patch(
            'enterprise_data.admin_analytics.utils.fetch_max_enrollment_datetime',
            return_value=datetime.now()
        )

        fetch_max_enrollment_datetime_patcher.start()
        self.addCleanup(fetch_max_enrollment_datetime_patcher.stop)

    def verify_engagement_data(self, results):
        """Verify the received engagement data."""
        attrs = [
            "email",
            "course_title",
            "activity_date",
            "course_subject",
        ]

        filtered_data = []
        for engagement in ENGAGEMENTS:
            for result in results:
                if engagement["email"] == result["email"]:
                    data = {attr: engagement[attr] for attr in attrs}
                    data["learning_time_hours"] = round(engagement["learning_time_seconds"] / 3600, 1)
                    filtered_data.append(data)
                    break

        received_data = sorted(results, key=lambda x: x["email"])
        expected_data = sorted(filtered_data, key=lambda x: x["email"])
        assert received_data == expected_data

    @patch(
        "enterprise_data.api.v1.views.analytics_engagements.fetch_and_cache_enrollments_data"
    )
    @patch(
        "enterprise_data.api.v1.views.analytics_engagements.fetch_and_cache_engagements_data"
    )
    def test_get(self, mock_fetch_and_cache_engagements_data, mock_fetch_and_cache_enrollments_data):
        """
        Test the GET method for the AdvanceAnalyticsEnrollmentStatsView works.
        """
        mock_fetch_and_cache_enrollments_data.return_value = enrollments_dataframe()
        mock_fetch_and_cache_engagements_data.return_value = engagements_dataframe()

        response = self.client.get(self.url)
        assert response.status_code == status.HTTP_200_OK
        data = response.json()
        assert data == {
            'engagements_over_time': [
                {'activity_date': '2021-07-19T00:00:00', 'enroll_type': 'certificate', 'sum': 0.0},
                {'activity_date': '2021-07-26T00:00:00', 'enroll_type': 'certificate', 'sum': 4.4},
                {'activity_date': '2021-07-27T00:00:00', 'enroll_type': 'certificate', 'sum': 1.2},
                {'activity_date': '2021-08-05T00:00:00', 'enroll_type': 'certificate', 'sum': 3.6},
                {'activity_date': '2021-08-21T00:00:00', 'enroll_type': 'certificate', 'sum': 2.7},
                {'activity_date': '2021-09-02T00:00:00', 'enroll_type': 'certificate', 'sum': 1.3},
                {'activity_date': '2021-09-21T00:00:00', 'enroll_type': 'certificate', 'sum': 1.5},
                {'activity_date': '2022-05-17T00:00:00', 'enroll_type': 'certificate', 'sum': 0.0}
            ],
            'top_courses_by_engagement': [
                {
                    'course_key': 'Kcpr+XoR30',
                    'course_title': 'Assimilated even-keeled focus group',
                    'enroll_type': 'certificate',
                    'count': 0.0
                },
                {
                    'course_key': 'luGg+KNt30',
                    'course_title': 'Synergized reciprocal encoding',
                    'enroll_type': 'certificate',
                    'count': 14.786944444444444
                }
            ],
            'top_subjects_by_engagement': [
                {
                    'course_subject': 'business-management',
                    'enroll_type': 'certificate',
                    'count': 14.786944444444444
                },
                {
                    'course_subject': 'engineering',
                    'enroll_type': 'certificate',
                    'count': 0.0
                }
            ]
        }

    @patch("enterprise_data.api.v1.views.analytics_engagements.fetch_and_cache_enrollments_data")
    @patch("enterprise_data.api.v1.views.analytics_engagements.fetch_and_cache_engagements_data")
    @ddt.data(
        EngagementChart.ENGAGEMENTS_OVER_TIME.value,
        EngagementChart.TOP_COURSES_BY_ENGAGEMENTS.value,
        EngagementChart.TOP_SUBJECTS_BY_ENGAGEMENTS.value,
    )
    def test_get_csv(self, chart_type, mock_fetch_and_cache_engagements_data, mock_fetch_and_cache_enrollments_data):
        """
        Test that AdvanceAnalyticsEngagementStatsView return correct CSV data.
        """
        mock_fetch_and_cache_enrollments_data.return_value = enrollments_dataframe()
        mock_fetch_and_cache_engagements_data.return_value = engagements_dataframe()
        response = self.client.get(self.url, {"response_type": ResponseType.CSV.value, "chart_type": chart_type})
        assert response.status_code == status.HTTP_200_OK
        assert response["Content-Type"] == "text/csv"
        # verify the response content
        assert response.content == ENGAGEMENT_STATS_CSVS[chart_type]

    @ddt.data(
        {
            "params": {"start_date": 1},
            "error": {
                "start_date": [
                    "Date has wrong format. Use one of these formats instead: YYYY-MM-DD."
                ]
            },
        },
        {
            "params": {"end_date": 2},
            "error": {
                "end_date": [
                    "Date has wrong format. Use one of these formats instead: YYYY-MM-DD."
                ]
            },
        },
        {
            "params": {"start_date": "2024-01-01", "end_date": "2023-01-01"},
            "error": {
                "non_field_errors": [
                    "start_date should be less than or equal to end_date."
                ]
            },
        },
        {
            "params": {"calculation": "invalid"},
            "error": {"calculation": [INVALID_CALCULATION_ERROR]},
        },
        {
            "params": {"granularity": "invalid"},
            "error": {"granularity": [INVALID_GRANULARITY_ERROR]},
        },
        {"params": {"chart_type": "invalid"}, "error": {"chart_type": [INVALID_CSV_ERROR1]}},
    )
    @ddt.unpack
    def test_get_invalid_query_params(self, params, error):
        """
        Test the GET method return correct error if any query param value is incorrect.
        """
        response = self.client.get(self.url, params)
        assert response.status_code == status.HTTP_400_BAD_REQUEST
        assert response.json() == error



"""Mock data for enrollments"""

import pandas as pd

from enterprise_data.admin_analytics.constants import EngagementChart, EnrollmentChart
from enterprise_data.admin_analytics.utils import ChartType

ENROLLMENTS = [
    {
        "enterprise_customer_name": "Hill Ltd",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "lms_enrollment_id": 1013,
        "user_id": 8907,
        "email": "rebeccanelson@example.com",
        "course_key": "hEmW+tvk03",
        "courserun_key": "course-v1:hEmW+tvk03+1T9889",
        "course_id": "1681",
        "course_subject": "business-management",
        "course_title": "Re-engineered tangible approach",
        "enterprise_enrollment_date": "2021-07-04",
        "lms_enrollment_mode": "verified",
        "enroll_type": "certificate",
        "program_title": "Non-Program",
        "date_certificate_awarded": "2021-08-25",
        "grade_percent": 0.99,
        "cert_awarded": 1,
        "date_certificate_created_raw": "2021-08-25",
        "passed_date_raw": "2021-08-25",
        "passed_date": "2021-08-25",
        "has_passed": 1,
    },
    {
        "enterprise_customer_name": "Hill Ltd",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "lms_enrollment_id": 9172,
        "user_id": 8369,
        "email": "taylorjames@example.com",
        "course_key": "hEmW+tvk03",
        "courserun_key": "course-v1:hEmW+tvk03+1T9889",
        "course_id": "1681",
        "course_subject": "business-management",
        "course_title": "Re-engineered tangible approach",
        "enterprise_enrollment_date": "2021-07-03",
        "lms_enrollment_mode": "verified",
        "enroll_type": "certificate",
        "program_title": "Non-Program",
        "date_certificate_awarded": "2021-09-01",
        "grade_percent": 0.93,
        "cert_awarded": 1,
        "date_certificate_created_raw": "2021-09-01",
        "passed_date_raw": "2021-09-01",
        "passed_date": "2021-09-01",
        "has_passed": 1,
    },
    {
        "enterprise_customer_name": "Hill Ltd",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "lms_enrollment_id": 9552,
        "user_id": 8719,
        "email": "ssmith@example.com",
        "course_key": "qZJC+KFX86",
        "courserun_key": "course-v1:qZJC+KFX86+1T8918",
        "course_id": "1725",
        "course_subject": "medicine",
        "course_title": "Secured static capability",
        "enterprise_enrollment_date": "2021-05-11",
        "lms_enrollment_mode": "verified",
        "enroll_type": "certificate",
        "program_title": "Non-Program",
        "date_certificate_awarded": None,
        "grade_percent": 0.0,
        "cert_awarded": 0,
        "date_certificate_created_raw": None,
        "passed_date_raw": None,
        "passed_date": None,
        "has_passed": 0,
    },
    {
        "enterprise_customer_name": "Hill Ltd",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "lms_enrollment_id": 3436,
        "user_id": 3125,
        "email": "kathleenmartin@example.com",
        "course_key": "QWXx+Jqz64",
        "courserun_key": "course-v1:QWXx+Jqz64+1T9449",
        "course_id": "4878",
        "course_subject": "social-sciences",
        "course_title": "Horizontal solution-oriented hub",
        "enterprise_enrollment_date": "2020-04-03",
        "lms_enrollment_mode": "verified",
        "enroll_type": "certificate",
        "program_title": "Non-Program",
        "date_certificate_awarded": None,
        "grade_percent": 0.0,
        "cert_awarded": 0,
        "date_certificate_created_raw": None,
        "passed_date_raw": None,
        "passed_date": None,
        "has_passed": 0,
    },
    {
        "enterprise_customer_name": "Hill Ltd",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "lms_enrollment_id": 5934,
        "user_id": 4853,
        "email": "amber79@example.com",
        "course_key": "NOGk+UVD31",
        "courserun_key": "course-v1:NOGk+UVD31+1T4956",
        "course_id": "4141",
        "course_subject": "communication",
        "course_title": "Streamlined zero-defect attitude",
        "enterprise_enrollment_date": "2020-04-08",
        "lms_enrollment_mode": "verified",
        "enroll_type": "certificate",
        "program_title": "Non-Program",
        "date_certificate_awarded": None,
        "grade_percent": 0.0,
        "cert_awarded": 0,
        "date_certificate_created_raw": None,
        "passed_date_raw": None,
        "passed_date": None,
        "has_passed": 0,
    },
]

ENGAGEMENTS = [
    {
        "user_id": 2340,
        "email": "padillamichelle@example.org",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-08-05",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 1,
        "is_engaged_video": 1,
        "is_engaged_forum": 0,
        "is_engaged_problem": 1,
        "is_active": 1,
        "learning_time_seconds": 3724,
    },
    {
        "user_id": 1869,
        "email": "yallison@example.org",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-07-27",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 1,
        "is_engaged_video": 1,
        "is_engaged_forum": 0,
        "is_engaged_problem": 1,
        "is_active": 1,
        "learning_time_seconds": 4335,
    },
    {
        "user_id": 6962,
        "email": "weaverpatricia@example.net",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-08-05",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 1,
        "is_engaged_video": 1,
        "is_engaged_forum": 0,
        "is_engaged_problem": 1,
        "is_active": 1,
        "learning_time_seconds": 9441,
    },
    {
        "user_id": 6798,
        "email": "seth57@example.org",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-08-21",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 1,
        "is_engaged_video": 1,
        "is_engaged_forum": 1,
        "is_engaged_problem": 1,
        "is_active": 1,
        "learning_time_seconds": 9898,
    },
    {
        "user_id": 8530,
        "email": "mackwilliam@example.com",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-10-24",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 0,
        "is_engaged_video": 0,
        "is_engaged_forum": 0,
        "is_engaged_problem": 0,
        "is_active": 1,
        "learning_time_seconds": 0,
    },
    {
        "user_id": 7584,
        "email": "graceperez@example.com",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2022-05-17",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 0,
        "is_engaged_video": 0,
        "is_engaged_forum": 0,
        "is_engaged_problem": 0,
        "is_active": 1,
        "learning_time_seconds": 21,
    },
    {
        "user_id": 68,
        "email": "yferguson@example.net",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-09-02",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 1,
        "is_engaged_video": 1,
        "is_engaged_forum": 0,
        "is_engaged_problem": 1,
        "is_active": 1,
        "learning_time_seconds": 4747,
    },
    {
        "user_id": 4278,
        "email": "caseyjohnny@example.com",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "Kcpr+XoR30",
        "enroll_type": "certificate",
        "activity_date": "2022-05-20",
        "course_title": "Assimilated even-keeled focus group",
        "course_subject": "engineering",
        "is_engaged": 0,
        "is_engaged_video": 0,
        "is_engaged_forum": 0,
        "is_engaged_problem": 0,
        "is_active": 1,
        "learning_time_seconds": 0,
    },
    {
        "user_id": 8726,
        "email": "webertodd@example.com",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-09-21",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 1,
        "is_engaged_video": 1,
        "is_engaged_forum": 1,
        "is_engaged_problem": 1,
        "is_active": 1,
        "learning_time_seconds": 5285,
    },
    {
        "user_id": 282,
        "email": "crystal86@example.net",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-07-11",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 0,
        "is_engaged_video": 0,
        "is_engaged_forum": 0,
        "is_engaged_problem": 0,
        "is_active": 1,
        "learning_time_seconds": 0,
    },
    {
        "user_id": 2731,
        "email": "paul77@example.org",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-07-26",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 1,
        "is_engaged_video": 1,
        "is_engaged_forum": 1,
        "is_engaged_problem": 1,
        "is_active": 1,
        "learning_time_seconds": 15753,
    },
    {
        "user_id": 4038,
        "email": "samanthaclarke@example.org",
        "enterprise_customer_uuid": "33ce656295e04ecfa2a77d407eb96f69",
        "course_key": "luGg+KNt30",
        "enroll_type": "certificate",
        "activity_date": "2021-07-19",
        "course_title": "Synergized reciprocal encoding",
        "course_subject": "business-management",
        "is_engaged": 0,
        "is_engaged_video": 0,
        "is_engaged_forum": 0,
        "is_engaged_problem": 0,
        "is_active": 1,
        "learning_time_seconds": 29,
    }
]

ENROLLMENT_STATS_CSVS = {
    EnrollmentChart.ENROLLMENTS_OVER_TIME.value: (
        b'enterprise_enrollment_date,certificate\n'
        b'2020-04-03,1\n'
        b'2020-04-08,1\n'
        b'2021-05-11,1\n'
        b'2021-07-03,1\n'
        b'2021-07-04,1\n'
    ),
    EnrollmentChart.TOP_COURSES_BY_ENROLLMENTS.value: (
        b'course_key,course_title,certificate\n'
        b'NOGk+UVD31,Streamlined zero-defect attitude,1\n'
        b'QWXx+Jqz64,Horizontal solution-oriented hub,1\n'
        b'hEmW+tvk03,Re-engineered tangible approach,2\n'
        b'qZJC+KFX86,Secured static capability,1\n'
    ),
    EnrollmentChart.TOP_SUBJECTS_BY_ENROLLMENTS.value: (
        b'course_subject,certificate\nbusiness-management,2\ncommunication,1\nmedicine,1\nsocial-sciences,1\n'
    )
}

ENGAGEMENT_STATS_CSVS = {
    EngagementChart.ENGAGEMENTS_OVER_TIME.value: (
        b'activity_date,certificate\n'
        b'2021-07-19,0.0\n'
        b'2021-07-26,4.4\n'
        b'2021-07-27,1.2\n'
        b'2021-08-05,3.6\n'
        b'2021-08-21,2.7\n'
        b'2021-09-02,1.3\n'
        b'2021-09-21,1.5\n'
        b'2022-05-17,0.0\n'
    ),
    EngagementChart.TOP_COURSES_BY_ENGAGEMENTS.value: (
        b'course_key,course_title,certificate\n'
        b'Kcpr+XoR30,Assimilated even-keeled focus group,0.0\n'
        b'luGg+KNt30,Synergized reciprocal encoding,14.786944444444444\n'
    ),
    EngagementChart.TOP_SUBJECTS_BY_ENGAGEMENTS.value: (
        b'course_subject,certificate\n'
        b'business-management,14.786944444444444\n'
        b'engineering,0.0\n'
    )
}


def enrollments_dataframe():
    """Return a DataFrame of enrollments."""
    enrollments = pd.DataFrame(ENROLLMENTS)

    enrollments['enterprise_enrollment_date'] = enrollments['enterprise_enrollment_date'].astype('datetime64[ns]')
    enrollments['date_certificate_awarded'] = enrollments['date_certificate_awarded'].astype('datetime64[ns]')
    enrollments['date_certificate_created_raw'] = enrollments['date_certificate_created_raw'].astype('datetime64[ns]')
    enrollments['passed_date_raw'] = enrollments['passed_date_raw'].astype('datetime64[ns]')
    enrollments['passed_date'] = enrollments['passed_date'].astype('datetime64[ns]')

    return enrollments


def engagements_dataframe():
    """Return a DataFrame of engagements."""
    engagements = pd.DataFrame(ENGAGEMENTS)
    engagements['activity_date'] = engagements['activity_date'].astype('datetime64[ns]')
    return engagements


def leaderboard_csv_content():
    """Return the CSV content of leaderboard."""
    # return (
    #     b'email,learning_time_hours,daily_sessions,average_session_length,course_completions\r\n'
    #     b'paul77@example.org,4.4,1,4.4,\r\nseth57@example.org,2.7,1,2.7,\r\n'
    #     b'weaverpatricia@example.net,2.6,1,2.6,\r\nwebertodd@example.com,1.5,1,1.5,\r\n'
    #     b'yferguson@example.net,1.3,1,1.3,\r\nyallison@example.org,1.2,1,1.2,\r\n'
    #     b'padillamichelle@example.org,1.0,1,1.0,\r\ncaseyjohnny@example.com,0.0,0,,\r\n'
    #     b'crystal86@example.net,0.0,0,,\r\ngraceperez@example.com,0.0,0,,\r\n'
    #     b'mackwilliam@example.com,0.0,0,,\r\nsamanthaclarke@example.org,0.0,0,,\r\n'
    # )

    return (
        b'email,learning_time_hours,daily_sessions,average_session_length,course_completions\r\n'
        b'paul77@example.org,4.4,1,4.4,\r\nseth57@example.org,2.7,1,2.7,\r\n'
        b'weaverpatricia@example.net,2.6,1,2.6,\r\nwebertodd@example.com,1.5,1,1.5,\r\n'
        b'yferguson@example.net,1.3,1,1.3,\r\nyallison@example.org,1.2,1,1.2,\r\n'
        b'padillamichelle@example.org,1.0,1,1.0,\r\ncaseyjohnny@example.com,0.0,0,0.0,\r\n'
        b'crystal86@example.net,0.0,0,0.0,\r\ngraceperez@example.com,0.0,0,0.0,\r\n'
        b'mackwilliam@example.com,0.0,0,0.0,\r\nsamanthaclarke@example.org,0.0,0,0.0,\r\n'
    )


LEADERBOARD_RESPONSE = [
    {
        "email": "paul77@example.org",
        "daily_sessions": 1,
        "learning_time_seconds": 15753,
        "learning_time_hours": 4.4,
        "average_session_length": 4.4,
        "course_completions": None,
    },
    {
        "email": "seth57@example.org",
        "daily_sessions": 1,
        "learning_time_seconds": 9898,
        "learning_time_hours": 2.7,
        "average_session_length": 2.7,
        "course_completions": None,
    },
    {
        "email": "weaverpatricia@example.net",
        "daily_sessions": 1,
        "learning_time_seconds": 9441,
        "learning_time_hours": 2.6,
        "average_session_length": 2.6,
        "course_completions": None,
    },
    {
        "email": "webertodd@example.com",
        "daily_sessions": 1,
        "learning_time_seconds": 5285,
        "learning_time_hours": 1.5,
        "average_session_length": 1.5,
        "course_completions": None,
    },
    {
        "email": "yferguson@example.net",
        "daily_sessions": 1,
        "learning_time_seconds": 4747,
        "learning_time_hours": 1.3,
        "average_session_length": 1.3,
        "course_completions": None,
    },
    {
        "email": "yallison@example.org",
        "daily_sessions": 1,
        "learning_time_seconds": 4335,
        "learning_time_hours": 1.2,
        "average_session_length": 1.2,
        "course_completions": None,
    },
    {
        "email": "padillamichelle@example.org",
        "daily_sessions": 1,
        "learning_time_seconds": 3724,
        "learning_time_hours": 1.0,
        "average_session_length": 1.0,
        "course_completions": None,
    },
    {
        "email": "caseyjohnny@example.com",
        "daily_sessions": 0,
        "learning_time_seconds": 0,
        "learning_time_hours": 0.0,
        "average_session_length": 0.0,
        "course_completions": None,
    },
    {
        "email": "crystal86@example.net",
        "daily_sessions": 0,
        "learning_time_seconds": 0,
        "learning_time_hours": 0.0,
        "average_session_length": 0.0,
        "course_completions": None,
    },
    {
        "email": "graceperez@example.com",
        "daily_sessions": 0,
        "learning_time_seconds": 21,
        "learning_time_hours": 0.0,
        "average_session_length": 0.0,
        "course_completions": None,
    },
    {
        "email": "mackwilliam@example.com",
        "daily_sessions": 0,
        "learning_time_seconds": 0,
        "learning_time_hours": 0.0,
        "average_session_length": 0.0,
        "course_completions": None,
    },
    {
        "email": "samanthaclarke@example.org",
        "daily_sessions": 0,
        "learning_time_seconds": 29,
        "learning_time_hours": 0.0,
        "average_session_length": 0.0,
        "course_completions": None,
    },
]

COMPLETIONS_STATS_CSVS = {
    ChartType.COMPLETIONS_OVER_TIME.value: (
        b'passed_date,certificate\n'
        b'2021-08-25,1\n'
        b'2021-09-01,2\n'
    ),
    ChartType.TOP_COURSES_BY_COMPLETIONS.value: (
        b'course_key,course_title,certificate\n'
        b'hEmW+tvk03,Re-engineered tangible approach,2\n'
    ),
    ChartType.TOP_SUBJECTS_BY_COMPLETIONS.value: (
        b'course_subject,certificate\n'
        b'business-management,2\n'
    )
}


def engagements_csv_content():
    """Return the CSV content of engagements."""
    return (
        b'email,course_title,activity_date,course_subject,learning_time_hours\r\n'
        b'graceperez@example.com,Synergized reciprocal encoding,2022-05-17,business-management,0.0\r\n'
        b'webertodd@example.com,Synergized reciprocal encoding,2021-09-21,business-management,1.5\r\n'
        b'yferguson@example.net,Synergized reciprocal encoding,2021-09-02,business-management,1.3\r\n'
        b'seth57@example.org,Synergized reciprocal encoding,2021-08-21,business-management,2.7\r\n'
        b'padillamichelle@example.org,Synergized reciprocal encoding,2021-08-05,business-management,1.0\r\n'
        b'weaverpatricia@example.net,Synergized reciprocal encoding,2021-08-05,business-management,2.6\r\n'
        b'yallison@example.org,Synergized reciprocal encoding,2021-07-27,business-management,1.2\r\n'
        b'paul77@example.org,Synergized reciprocal encoding,2021-07-26,business-management,4.4\r\n'
        b'samanthaclarke@example.org,Synergized reciprocal encoding,2021-07-19,business-management,0.0\r\n'
    )



"""
Test the utility functions in the admin_analytics app.
"""
from datetime import datetime, timedelta

import pandas
from mock import patch

from django.test import TestCase

from enterprise_data.admin_analytics.utils import (
    fetch_and_cache_engagements_data,
    fetch_and_cache_enrollments_data,
    fetch_and_cache_skills_data,
    get_cache_timeout,
    get_skills_bubble_chart_df,
    get_top_skills_completion,
    get_top_skills_enrollment,
)
from enterprise_data.utils import date_filter


class TestUtils(TestCase):
    """
    Test suite for the utility functions in the admin_analytics package.
    """

    def test_get_cache_timeout(self):
        """
        Validate the get_cache_timeout function.
        """
        now = datetime.now().replace(microsecond=0)
        with patch('enterprise_data.admin_analytics.utils.datetime') as mock_datetime:
            mock_datetime.now.return_value = now
            cache_expiry = now
            self.assertEqual(get_cache_timeout(cache_expiry), 0)

            cache_expiry = now + timedelta(seconds=10)
            self.assertEqual(get_cache_timeout(cache_expiry), 10)

            cache_expiry = now + timedelta(seconds=100)
            self.assertEqual(get_cache_timeout(cache_expiry), 100)

            # Validate the case where cache_expiry is in the past.
            cache_expiry = now - timedelta(seconds=10)
            self.assertEqual(get_cache_timeout(cache_expiry), 0)

    def test_fetch_and_cache_enrollments_data(self):
        """
        Validate the fetch_and_cache_enrollments_data function.
        """
        with patch('enterprise_data.admin_analytics.utils.fetch_enrollment_data') as mock_fetch_enrollment_data:
            with patch('enterprise_data.admin_analytics.utils.TieredCache') as mock_tiered_cache:
                # Simulate the scenario where the data is not found in the cache.
                mock_tiered_cache.get_cached_response.return_value.is_found = False
                mock_fetch_enrollment_data.return_value = 'enrollments'

                enrollments = fetch_and_cache_enrollments_data('enterprise_id', datetime.now() + timedelta(seconds=10))
                self.assertEqual(enrollments, 'enrollments')
                self.assertEqual(mock_tiered_cache.get_cached_response.call_count, 1)
                self.assertEqual(mock_tiered_cache.set_all_tiers.call_count, 1)

    def test_fetch_and_cache_enrollments_data_with_data_cache_found(self):
        """
        Validate the fetch_and_cache_enrollments_data function.
        """
        with patch('enterprise_data.admin_analytics.utils.fetch_enrollment_data') as mock_fetch_enrollment_data:
            with patch('enterprise_data.admin_analytics.utils.TieredCache') as mock_tiered_cache:
                # Simulate the scenario where the data is found in the cache.
                mock_tiered_cache.get_cached_response.return_value.is_found = True
                mock_tiered_cache.get_cached_response.return_value.value = 'cached-enrollments'
                mock_fetch_enrollment_data.return_value = 'enrollments'

                enrollments = fetch_and_cache_enrollments_data('enterprise_id', datetime.now() + timedelta(seconds=10))
                self.assertEqual(enrollments, 'cached-enrollments')
                self.assertEqual(mock_tiered_cache.get_cached_response.call_count, 1)
                self.assertEqual(mock_tiered_cache.set_all_tiers.call_count, 0)

    def test_fetch_and_cache_engagements_data(self):
        """
        Validate the fetch_and_cache_engagements_data function.
        """
        with patch('enterprise_data.admin_analytics.utils.fetch_engagement_data') as mock_fetch_engagement_data:
            with patch('enterprise_data.admin_analytics.utils.TieredCache') as mock_tiered_cache:
                # Simulate the scenario where the data is not found in the cache.
                mock_tiered_cache.get_cached_response.return_value.is_found = False
                mock_fetch_engagement_data.return_value = 'engagements'

                enrollments = fetch_and_cache_engagements_data('enterprise_id', datetime.now() + timedelta(seconds=10))
                self.assertEqual(enrollments, 'engagements')
                self.assertEqual(mock_tiered_cache.get_cached_response.call_count, 1)
                self.assertEqual(mock_tiered_cache.set_all_tiers.call_count, 1)

    def test_fetch_and_cache_engagements_data_with_data_cache_found(self):
        """
        Validate the fetch_and_cache_engagements_data function.
        """
        with patch('enterprise_data.admin_analytics.utils.fetch_engagement_data') as mock_fetch_engagement_data:
            with patch('enterprise_data.admin_analytics.utils.TieredCache') as mock_tiered_cache:
                # Simulate the scenario where the data is found in the cache.
                mock_tiered_cache.get_cached_response.return_value.is_found = True
                mock_tiered_cache.get_cached_response.return_value.value = 'cached-engagements'
                mock_fetch_engagement_data.return_value = 'engagements'

                enrollments = fetch_and_cache_engagements_data('enterprise_id', datetime.now() + timedelta(seconds=10))
                self.assertEqual(enrollments, 'cached-engagements')
                self.assertEqual(mock_tiered_cache.get_cached_response.call_count, 1)
                self.assertEqual(mock_tiered_cache.set_all_tiers.call_count, 0)

    def test_fetch_and_cache_skills_data(self):
        """
        Validate the fetch_and_cache_skills_data function.
        """
        with patch('enterprise_data.admin_analytics.utils.fetch_skills_data') as mock_fetch_skills_data:
            with patch('enterprise_data.admin_analytics.utils.TieredCache') as mock_tiered_cache:
                # Simulate the scenario where the data is not found in the cache.
                mock_tiered_cache.get_cached_response.return_value.is_found = False
                mock_fetch_skills_data.return_value = 'skills'
                skills = fetch_and_cache_skills_data('enterprise_id', datetime.now() + timedelta(seconds=10))
                self.assertEqual(skills, 'skills')
                self.assertEqual(mock_tiered_cache.get_cached_response.call_count, 1)
                self.assertEqual(mock_tiered_cache.set_all_tiers.call_count, 1)

    def test_fetch_and_cache_skills_data_with_data_cache_found(self):
        """
        Validate the fetch_and_cache_skills_data function.
        """
        with patch('enterprise_data.admin_analytics.utils.fetch_skills_data') as mock_fetch_skills_data:
            with patch('enterprise_data.admin_analytics.utils.TieredCache') as mock_tiered_cache:
                # Simulate the scenario where the data is found in the cache.
                mock_tiered_cache.get_cached_response.return_value.is_found = True
                mock_tiered_cache.get_cached_response.return_value.value = 'cached-skills'
                mock_fetch_skills_data.return_value = 'skills'

                skills = fetch_and_cache_skills_data('enterprise_id', datetime.now() + timedelta(seconds=10))
                self.assertEqual(skills, 'cached-skills')
                self.assertEqual(mock_tiered_cache.get_cached_response.call_count, 1)
                self.assertEqual(mock_tiered_cache.set_all_tiers.call_count, 0)

    def test_get_skills_bubble_chart_df(self):
        """
        Validate the get_skills_bubble_chart_df function.
        """
        # Mock skills data
        skills = pandas.DataFrame({
            'skill_name': ['Skill A', 'Skill B', 'Skill C'],
            'skill_type': ['Type 1', 'Type 2', 'Type 1'],
            'enrolls': [100, 200, 150],
            'completions': [50, 100, 75],
            'date': [datetime.now(), datetime.now(), datetime.now()],
        })

        # Define the expected result
        expected_result = pandas.DataFrame({
            'skill_name': ['Skill B', 'Skill C', 'Skill A'],
            'skill_type': ['Type 2', 'Type 1', 'Type 1'],
            'enrolls': [200, 150, 100],
            'completions': [100, 75, 50]
        })
        # Reset the index for the expected result
        expected_result = expected_result.reset_index(drop=True)

        start_date = datetime.now() - timedelta(days=10)
        end_date = datetime.now()
        filtered_skills = date_filter(start_date, end_date, skills, 'date')
        # Call the function
        result = get_skills_bubble_chart_df(filtered_skills)

        # Reset the index for the result
        result = result.reset_index(drop=True)

        # Assert the result
        pandas.testing.assert_frame_equal(result, expected_result)

    def test_get_top_skills_enrollment(self):
        """
        Validate the get_top_skills_enrollment function.
        """
        # Mock skills data
        skills = pandas.DataFrame({
            'primary_subject_name': ['engineering', 'communication', 'engineering', 'other'],
            'skill_name': ['Skill A', 'Skill B', 'Skill A', 'Skill D'],
            'enrolls': [100, 200, 150, 300],
            'completions': [50, 100, 75, 150],
            'date': [datetime.now(), datetime.now(), datetime.now(), datetime.now()],
        })

        # Define the expected result
        expected_result = pandas.DataFrame({
            'skill_name': ['Skill B', 'Skill A', 'Skill D'],
            'primary_subject_name': ['communication', 'engineering', 'other'],
            'count': [200, 250, 300]
        })
        # Reset the index for the expected result
        expected_result = expected_result.reset_index(drop=True)
        start_date = datetime.now() - timedelta(days=10)
        end_date = datetime.now()
        filtered_skills = date_filter(start_date, end_date, skills, 'date')

        # Call the function
        result = get_top_skills_enrollment(filtered_skills)

        # Reset the index for the result
        result = result.reset_index(drop=True)

        # Assert the result
        pandas.testing.assert_frame_equal(result, expected_result)

    def test_get_top_skills_completion(self):
        """
        Validate the get_top_skills_completion function.
        """
        # Mock skills data
        skills = pandas.DataFrame({
            'skill_name': ['Skill A', 'Skill B', 'Skill C', 'Skill D'],
            'primary_subject_name': ['communication', 'engineering', 'communication', 'other'],
            'enrolls': [100, 200, 150, 300],
            'completions': [50, 100, 75, 150],
            'date': [datetime.now(), datetime.now(), datetime.now(), datetime.now()]
        })

        # Define the expected result
        expected_result = pandas.DataFrame({
            'skill_name': ['Skill A', 'Skill C', 'Skill B', 'Skill D'],
            'primary_subject_name': ['communication', 'communication', 'engineering', 'other'],
            'count': [50, 75, 100, 150]
        })
        # Reset the index for the expected result
        expected_result = expected_result.reset_index(drop=True)
        start_date = datetime.now() - timedelta(days=10)
        end_date = datetime.now()
        filtered_skills = date_filter(start_date, end_date, skills, 'date')

        # Call the function
        result = get_top_skills_completion(filtered_skills)

        # Reset the index for the result
        result = result.reset_index(drop=True)

        # Assert the result
        pandas.testing.assert_frame_equal(result, expected_result)



"""
Utility functions for fetching data from the database.
"""
from datetime import datetime, timedelta
from enum import Enum
from logging import getLogger

from edx_django_utils.cache import TieredCache, get_cache_key

from enterprise_data.admin_analytics.constants import Calculation, Granularity
from enterprise_data.admin_analytics.data_loaders import (
    fetch_engagement_data,
    fetch_enrollment_data,
    fetch_max_enrollment_datetime,
    fetch_skills_data,
)
from enterprise_data.utils import date_filter, primary_subject_truncate

LOGGER = getLogger(__name__)


class ChartType(Enum):
    """
    Chart types.
    """
    BUBBLE = 'bubble'
    TOP_SKILLS_ENROLLMENT = 'top_skills_enrollment'
    TOP_SKILLS_COMPLETION = 'top_skills_completion'
    COMPLETIONS_OVER_TIME = 'completions_over_time'
    TOP_COURSES_BY_COMPLETIONS = 'top_courses_by_completions'
    TOP_SUBJECTS_BY_COMPLETIONS = 'top_subjects_by_completions'


def fetch_enrollments_cache_expiry_timestamp():
    """Calculate cache expiry timestamp"""
    # TODO: Implement correct cache expiry logic for `enrollments` data.
    #       Current cache expiry logic is based on `enterprise_learner_enrollment` table,
    #       Which has nothing to do with the `enrollments` data. Instead cache expiry should
    #       be based on `fact_enrollment_admin_dash` table. Currently we have no timestamp in
    #       `fact_enrollment_admin_dash` table that can be used for cache expiry. Add a new
    #       column in the table for this purpose and then use that column for cache expiry.
    last_updated_at = fetch_max_enrollment_datetime()
    cache_expiry = (
        last_updated_at + timedelta(days=1) if last_updated_at else datetime.now()
    )
    return cache_expiry


def fetch_engagements_cache_expiry_timestamp():
    """Calculate cache expiry timestamp"""
    # TODO: Implement correct cache expiry logic for `engagements` data.
    #       Current cache expiry logic is based on `enterprise_learner_enrollment` table,
    #       Which has nothing to do with the `engagements` data. Instead cache expiry should
    #       be based on `fact_enrollment_engagement_day_admin_dash` table. Currently we have
    #       no timestamp in `fact_enrollment_engagement_day_admin_dash` table that can be used
    #       for cache expiry. Add a new column in the table for this purpose and then use that
    #       column for cache expiry.
    last_updated_at = fetch_max_enrollment_datetime()
    cache_expiry = (
        last_updated_at + timedelta(days=1) if last_updated_at else datetime.now()
    )
    return cache_expiry


def granularity_aggregation(level, group, date, data_frame, aggregation_type="count"):
    """Aggregate data based on granularity"""
    df = data_frame

    period_mapping = {
        Granularity.WEEKLY.value: "W",
        Granularity.MONTHLY.value: "M",
        Granularity.QUARTERLY.value: "Q"
    }

    if level in period_mapping:
        df[date] = df[date].dt.to_period(period_mapping[level]).dt.start_time

    agg_column_name = "count"
    if aggregation_type == "count":
        df = df.groupby(group).size().reset_index()
    elif aggregation_type == "sum":
        df = df.groupby(group).sum().reset_index()
        agg_column_name = "sum"

    df.columns = group + [agg_column_name]
    return df


def calculation_aggregation(calc, data_frame, aggregation_type="count"):
    """Aggregate data based on calculation"""
    df = data_frame

    window_mapping = {
        Calculation.MOVING_AVERAGE_3_PERIOD.value: 3,
        Calculation.MOVING_AVERAGE_7_PERIOD.value: 7,
    }

    aggregation_column = "count" if aggregation_type == "count" else "sum"

    if calc == Calculation.RUNNING_TOTAL.value:
        df[aggregation_column] = df.groupby("enroll_type")[aggregation_column].cumsum()
    elif calc in [Calculation.MOVING_AVERAGE_3_PERIOD.value, Calculation.MOVING_AVERAGE_7_PERIOD.value]:
        df[aggregation_column] = (
            df.groupby("enroll_type")[aggregation_column]
            .rolling(window_mapping[calc])
            .mean()
            .droplevel(level=[0])
        )

    return df


def get_cache_timeout(cache_expiry):
    """
    Helper method to calculate cache timeout in seconds.

    Arguments:
        cache_expiry (datetime): Datetime object denoting the cache expiry.

    Returns:
        (int): Cache timeout in seconds.
    """
    now = datetime.now()
    cache_timeout = 0
    if cache_expiry > now:
        # Calculate cache expiry in seconds from now.
        cache_timeout = (cache_expiry - now).seconds

    return cache_timeout


def fetch_and_cache_enrollments_data(enterprise_id, cache_expiry):
    """
    Helper method to fetch and cache enrollments data.

    Arguments:
        enterprise_id (str): UUID of the enterprise customer in string format.
        cache_expiry (datetime): Datetime object denoting the cache expiry.

    Returns:
        (pandas.DataFrame): The enrollments data.
    """
    cache_key = get_cache_key(
        resource='enterprise-admin-analytics-aggregates-enrollments',
        enterprise_customer=enterprise_id,
    )
    cached_response = TieredCache.get_cached_response(cache_key)

    if cached_response.is_found:
        LOGGER.info(f"Enrollments data found in cache for Enterprise [{enterprise_id}]")
        return cached_response.value
    else:
        enrollments = fetch_enrollment_data(enterprise_id)
        TieredCache.set_all_tiers(
            cache_key, enrollments, get_cache_timeout(cache_expiry)
        )
        return enrollments


def fetch_and_cache_engagements_data(enterprise_id, cache_expiry):
    """
    Helper method to fetch and cache engagements data.

    Arguments:
        enterprise_id (str): UUID of the enterprise customer in string format.
        cache_expiry (datetime): Datetime object denoting the cache expiry.

    Returns:
        (pandas.DataFrame): The engagements data.
    """
    cache_key = get_cache_key(
        resource='enterprise-admin-analytics-aggregates-engagements',
        enterprise_customer=enterprise_id,
    )
    cached_response = TieredCache.get_cached_response(cache_key)

    if cached_response.is_found:
        LOGGER.info(f"Engagements data found in cache for Enterprise [{enterprise_id}]")
        return cached_response.value
    else:
        engagements = fetch_engagement_data(enterprise_id)
        TieredCache.set_all_tiers(
            cache_key, engagements, get_cache_timeout(cache_expiry)
        )
        return engagements


def fetch_and_cache_skills_data(enterprise_id, cache_expiry):
    """
    Helper method to fetch and cache skills data.

    Arguments:
        enterprise_id (str): UUID of the enterprise customer in string format.
        cache_expiry (datetime): Datetime object denoting the cache expiry.

    Returns:
        (pandas.DataFrame): The skills data.
    """
    cache_key = get_cache_key(
        resource='enterprise-admin-analytics-aggregate-skills',
        enterprise_customer=enterprise_id,
    )
    cached_response = TieredCache.get_cached_response(cache_key)

    if cached_response.is_found:
        LOGGER.info(f"Skills data found in cache for Enterprise [{enterprise_id}]")
        return cached_response.value
    else:
        skills = fetch_skills_data(enterprise_id)
        TieredCache.set_all_tiers(
            cache_key, skills, get_cache_timeout(cache_expiry)
        )
        return skills


def get_skills_bubble_chart_df(skills_filtered):
    """ Get the skills data for the bubble chart.

    Args:
        skills_filtered (pandas.DataFrame): The skills data.

    Returns:
        (pandas.DataFrame): The skills data for the bubble chart.
    """

    # Group by skill_name and skill_type, and aggregate enrolls and completions
    skills_aggregated = (
        skills_filtered.groupby(['skill_name', 'skill_type'], as_index=False)
        .agg(enrolls=('enrolls', 'sum'), completions=('completions', 'sum'))
    )

    # Convert enrolls and completions to integers
    skills_aggregated['enrolls'] = skills_aggregated['enrolls'].astype(int)
    skills_aggregated['completions'] = skills_aggregated['completions'].astype(int)

    # Sort the dataframe by enrolls and completions in descending order
    skills_aggregated = skills_aggregated.sort_values(by=['enrolls', 'completions'], ascending=False)

    return skills_aggregated


def get_top_skills_enrollment(skills_filtered):
    """ Get the top skills by enrolls.

    Args:
        skills_filtered (pandas.DataFrame): The skills data.

    Returns:
        (pandas.DataFrame): The top skills by enrolls data
    """

    # Get the top 10 skills by enrolls
    top_skills = (
        skills_filtered.groupby('skill_name')
        .enrolls.sum()
        .sort_values(ascending=False)
        .head(10)
        .index
    )

    # Apply primary_subject_truncate to the primary_subject_name column
    skills_filtered['primary_subject_name'] = skills_filtered['primary_subject_name'].apply(primary_subject_truncate)

    # Filter data for the top skills and aggregate enrolls by skill_name and primary_subject_name
    top_skills_enrollment_data = (
        skills_filtered[skills_filtered.skill_name.isin(top_skills)]
        .groupby(['skill_name', 'primary_subject_name'], as_index=False)
        .agg(count=('enrolls', 'sum'))
    )

    # Sort the dataframe by primary_subject_name
    top_skills_enrollment_data = top_skills_enrollment_data.sort_values(by="primary_subject_name")

    return top_skills_enrollment_data


def get_top_skills_completion(skills_filtered):
    """ Get the top skills by completions.

    Args:
        skills_filtered (pandas.DataFrame): The skills data.

    Returns:
        (pandas.DataFrame): The top skills by completions
    """

    # Get the top 10 skills by completions
    top_skills = (
        skills_filtered.groupby('skill_name')
        .completions.sum()
        .sort_values(ascending=False)
        .head(10)
        .index
    )

    # Apply primary_subject_truncate to the primary_subject_name column
    skills_filtered['primary_subject_name'] = skills_filtered['primary_subject_name'].apply(primary_subject_truncate)

    # Filter data for the top skills and aggregate completions by skill_name and primary_subject_name
    top_skills_completion_data = (
        skills_filtered[skills_filtered.skill_name.isin(top_skills)]
        .groupby(['skill_name', 'primary_subject_name'], as_index=False)
        .agg(count=('completions', 'sum'))
    )

    # Sort the dataframe by primary_subject_name
    top_skills_completion_data = top_skills_completion_data.sort_values(by='primary_subject_name')

    return top_skills_completion_data


def get_skills_chart_data(chart_type, start_date, end_date, skills):
    """
    Get chart data for skill charts.

    Arguments:
        chart_type (ChartType): The type of chart to generate.
        start_date (datetime): The start date for the date filter.
        end_date (datetime): The end date for the date filter.
        skills (pandas.DataFrame): The skills data.
    """
    skills_filtered = date_filter(start=start_date, end=end_date, data_frame=skills.copy(), date_column='date')
    if chart_type == ChartType.BUBBLE:
        return get_skills_bubble_chart_df(skills_filtered=skills_filtered.copy())
    elif chart_type == ChartType.TOP_SKILLS_ENROLLMENT:
        return get_top_skills_enrollment(skills_filtered=skills_filtered.copy())
    elif chart_type == ChartType.TOP_SKILLS_COMPLETION:
        return get_top_skills_completion(skills_filtered=skills_filtered.copy())
    else:
        raise ValueError(f"Invalid chart type: {chart_type}")


def get_top_skills_csv_data(skills, start_date, end_date):
    """ Get the top skills data for CSV download.

    Args:
        skills (pandas.DataFrame): The skills data.
        start_date (str): The start date for the date filter.
        end_date (str): The end date for the date filter.

    Returns:
        (pandas.DataFrame): The top skills data for CSV download.
    """
    dff = get_skills_chart_data(
        chart_type=ChartType.BUBBLE,
        start_date=start_date,
        end_date=end_date,
        skills=skills.copy(),
    )
    dff = dff.sort_values(by='enrolls', ascending=False)
    return dff



"""Advance Analytics for Leaderboard"""
from datetime import datetime
from logging import getLogger

import numpy as np
import pandas as pd
from edx_rbac.decorators import permission_required
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from rest_framework.views import APIView

from django.http import StreamingHttpResponse

from enterprise_data.admin_analytics.constants import ResponseType
from enterprise_data.admin_analytics.utils import (
    fetch_and_cache_engagements_data,
    fetch_and_cache_enrollments_data,
    fetch_engagements_cache_expiry_timestamp,
    fetch_enrollments_cache_expiry_timestamp,
)
from enterprise_data.api.v1.paginators import AdvanceAnalyticsPagination
from enterprise_data.api.v1.serializers import AdvanceAnalyticsQueryParamSerializer
from enterprise_data.renderers import LeaderboardCSVRenderer
from enterprise_data.utils import date_filter

LOGGER = getLogger(__name__)


class AdvanceAnalyticsLeaderboardView(APIView):
    """
    API for getting the advance analytics leaderboard data.
    """
    authentication_classes = (JwtAuthentication,)
    pagination_class = AdvanceAnalyticsPagination
    http_method_names = ['get']

    @permission_required('can_access_enterprise', fn=lambda request, enterprise_uuid: enterprise_uuid)
    def get(self, request, enterprise_uuid):
        """Get leaderboard data"""
        serializer = AdvanceAnalyticsQueryParamSerializer(data=request.GET)
        serializer.is_valid(raise_exception=True)

        enrollments_cache_expiry = fetch_enrollments_cache_expiry_timestamp()
        enrollments_df = fetch_and_cache_enrollments_data(enterprise_uuid, enrollments_cache_expiry)

        engagements_cache_expiry = fetch_engagements_cache_expiry_timestamp()
        engagements_df = fetch_and_cache_engagements_data(enterprise_uuid, engagements_cache_expiry)

        start_date = serializer.data.get('start_date', enrollments_df.enterprise_enrollment_date.min())
        end_date = serializer.data.get('end_date', datetime.now())
        response_type = serializer.data.get('response_type', ResponseType.JSON.value)

        LOGGER.info(
            "Leaderboard data requested for enterprise [%s] from [%s] to [%s]",
            enterprise_uuid,
            start_date,
            end_date,
        )

        # only include learners who have passed the course
        enrollments_df = enrollments_df[enrollments_df["has_passed"] == 1]

        # filter enrollments by date
        enrollments_df = date_filter(start_date, end_date, enrollments_df, "passed_date")

        completions = enrollments_df.groupby(["email"]).size().reset_index()
        completions.columns = ["email", "course_completions"]

        # filter engagements by date
        engagements_df = date_filter(start_date, end_date, engagements_df, "activity_date")

        engage = (
            engagements_df.groupby(["email"])
            .agg({"is_engaged": ["sum"], "learning_time_seconds": ["sum"]})
            .reset_index()
        )
        engage.columns = ["email", "daily_sessions", "learning_time_seconds"]
        engage["learning_time_hours"] = round(
            engage["learning_time_seconds"].astype("float") / 60 / 60, 1
        )

        # if daily_sessions is 0, set average_session_length to 0 becuase otherwise it will be `inf`
        engage["average_session_length"] = np.where(
            engage["daily_sessions"] == 0,
            0,
            round(engage["learning_time_hours"] / engage["daily_sessions"].astype("float"), 1)
        )

        leaderboard_df = engage.merge(completions, on="email", how="left")
        leaderboard_df = leaderboard_df.sort_values(
            by=["learning_time_hours", "daily_sessions", "course_completions"],
            ascending=[False, False, False],
        )

        # move the aggregated row with email 'null' to the end of the table
        idx = leaderboard_df.index[leaderboard_df['email'] == 'null']
        leaderboard_df.loc[idx, 'email'] = 'learners who have not shared consent'
        leaderboard_df = pd.concat([leaderboard_df.drop(idx), leaderboard_df.loc[idx]])

        # convert `nan` values to `None` because `nan` is not JSON serializable
        leaderboard_df = leaderboard_df.replace(np.nan, None)

        LOGGER.info(
            "Leaderboard data prepared for enterprise [%s] from [%s] to [%s]",
            enterprise_uuid,
            start_date,
            end_date,
        )

        if response_type == ResponseType.CSV.value:
            filename = f"""Leaderboard, {start_date} - {end_date}.csv"""
            leaderboard_df = leaderboard_df[
                [
                    "email",
                    "learning_time_hours",
                    "daily_sessions",
                    "average_session_length",
                    "course_completions",
                ]
            ]
            return StreamingHttpResponse(
                LeaderboardCSVRenderer().render(self._stream_serialized_data(leaderboard_df)),
                content_type="text/csv",
                headers={
                    "Content-Disposition": f'attachment; filename="{filename}"',
                    "Access-Control-Expose-Headers": "Content-Disposition"
                },
            )

        paginator = self.pagination_class()
        page = paginator.paginate_queryset(leaderboard_df, request)
        serialized_data = page.data.to_dict(orient='records')
        response = paginator.get_paginated_response(serialized_data)

        return response

    def _stream_serialized_data(self, leaderboard_df, chunk_size=50000):
        """
        Stream the serialized data.
        """
        total_rows = leaderboard_df.shape[0]
        for start_index in range(0, total_rows, chunk_size):
            end_index = min(start_index + chunk_size, total_rows)
            chunk = leaderboard_df.iloc[start_index:end_index]
            yield from chunk.to_dict(orient='records')



"""
Module for interacting with the fact_enrollment_engagement_day_admin_dash table.
"""
from datetime import date
from uuid import UUID

from ..queries import FactEngagementAdminDashQueries
from ..utils import run_query
from .base import BaseTable


class FactEngagementAdminDashTable(BaseTable):
    """
    Class for communicating with the fact_enrollment_engagement_day_admin_dash table.
    """
    queries = FactEngagementAdminDashQueries()

    def get_learning_hours_and_daily_sessions(self, enterprise_customer_uuid: UUID, start_date: date, end_date: date):
        """
        Get the learning hours and daily sessions for the given enterprise customer.

        Arguments:
            enterprise_customer_uuid (UUID): The UUID of the enterprise customer.
            start_date (date): The start date.
            end_date (date): The end date.

        Returns:
            (tuple<float, int>): The learning hours and daily sessions.
        """
        results = run_query(
            query=self.queries.get_learning_hours_and_daily_sessions_query(),
            params={
                'enterprise_customer_uuid': enterprise_customer_uuid,
                'start_date': start_date,
                'end_date': end_date,
            }
        )
        return tuple(results[0])



"""Advanced Analytics Constants"""

from enum import Enum


class Granularity(Enum):
    """Granularity choices"""
    DAILY = 'Daily'
    WEEKLY = 'Weekly'
    MONTHLY = 'Monthly'
    QUARTERLY = 'Quarterly'


class Calculation(Enum):
    """Calculation choices"""
    TOTAL = 'Total'
    RUNNING_TOTAL = 'Running Total'
    MOVING_AVERAGE_3_PERIOD = 'Moving Average (3 Period)'
    MOVING_AVERAGE_7_PERIOD = 'Moving Average (7 Period)'


class EnrollmentChart(Enum):
    """CSV choices"""
    ENROLLMENTS_OVER_TIME = 'enrollments_over_time'
    TOP_COURSES_BY_ENROLLMENTS = 'top_courses_by_enrollments'
    TOP_SUBJECTS_BY_ENROLLMENTS = 'top_subjects_by_enrollments'
    INDIVIDUAL_ENROLLMENTS = 'individual_enrollments'


class ResponseType(Enum):
    """Response type choices"""
    JSON = 'json'
    CSV = 'csv'


class EngagementChart(Enum):
    """Response Choices"""
    ENGAGEMENTS_OVER_TIME = 'engagements_over_time'
    TOP_COURSES_BY_ENGAGEMENTS = 'top_courses_by_engagements'
    TOP_SUBJECTS_BY_ENGAGEMENTS = 'top_subjects_by_engagements'
    INDIVIDUAL_ENGAGEMENTS = 'individual_engagements'



"""Unitest for EnterpriseAdminCompletionsStatsView."""
from datetime import datetime

import ddt
from mock import patch
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITransactionTestCase

from enterprise_data.admin_analytics.utils import ChartType
from enterprise_data.tests.admin_analytics.mock_analytics_data import (
    COMPLETIONS_STATS_CSVS,
    ENROLLMENTS,
    enrollments_dataframe,
)
from enterprise_data.tests.mixins import JWTTestMixin
from enterprise_data.tests.test_utils import UserFactory
from enterprise_data_roles.constants import ENTERPRISE_DATA_ADMIN_ROLE
from enterprise_data_roles.models import EnterpriseDataFeatureRole, EnterpriseDataRoleAssignment


@ddt.ddt
class TestCompletionstStatsAPI(JWTTestMixin, APITransactionTestCase):
    """Tests for EnterrpiseAdminCompletionsStatsView."""

    def setUp(self):
        """
        Setup method.
        """
        super().setUp()
        self.user = UserFactory(is_staff=True)
        role, __ = EnterpriseDataFeatureRole.objects.get_or_create(
            name=ENTERPRISE_DATA_ADMIN_ROLE
        )
        self.role_assignment = EnterpriseDataRoleAssignment.objects.create(
            role=role, user=self.user
        )
        self.client.force_authenticate(user=self.user)

        self.enterprise_id = "ee5e6b3a-069a-4947-bb8d-d2dbc323396c"
        self.set_jwt_cookie()

        self.url = reverse(
            "v1:enterprise-admin-analytics-completions-stats",
            kwargs={"enterprise_id": self.enterprise_id},
        )

        fetch_max_enrollment_datetime_patcher = patch(
            'enterprise_data.api.v1.views.enterprise_completions.fetch_max_enrollment_datetime',
            return_value=datetime.now()
        )

        fetch_max_enrollment_datetime_patcher.start()
        self.addCleanup(fetch_max_enrollment_datetime_patcher.stop)

    @patch(
        "enterprise_data.api.v1.views.enterprise_completions.fetch_and_cache_enrollments_data"
    )
    def test_get(self, mock_fetch_and_cache_enrollments_data):
        """
        Test the GET method for the EnterrpiseAdminCompletionsStatsView works.
        """
        mock_fetch_and_cache_enrollments_data.return_value = enrollments_dataframe()

        params = {
            "start_date": "2020-01-01",
            "end_date": "2025-08-09",
            "calculation": "Running Total",
            "granularity": "Daily",
        }
        response = self.client.get(self.url, params)
        assert response.status_code == status.HTTP_200_OK
        data = response.json()
        assert data == {
            "completions_over_time": [
                {
                    "passed_date": "2021-08-25T00:00:00",
                    "enroll_type": "certificate",
                    "count": 1,
                },
                {
                    "passed_date": "2021-09-01T00:00:00",
                    "enroll_type": "certificate",
                    "count": 2,
                },
            ],
            "top_courses_by_completions": [
                {
                    "course_key": "hEmW+tvk03",
                    "course_title": "Re-engineered tangible approach",
                    "enroll_type": "certificate",
                    "count": 2,
                }
            ],
            "top_subjects_by_completions": [
                {
                    "course_subject": "business-management",
                    "enroll_type": "certificate",
                    "count": 2,
                }
            ],
        }

    @patch("enterprise_data.api.v1.views.enterprise_completions.fetch_and_cache_enrollments_data")
    @ddt.data(
        ChartType.COMPLETIONS_OVER_TIME.value,
        ChartType.TOP_COURSES_BY_COMPLETIONS.value,
        ChartType.TOP_SUBJECTS_BY_COMPLETIONS.value,
    )
    def test_get_csv(self, chart_type, mock_fetch_and_cache_enrollments_data):
        """
        Test that EnterrpiseAdminCompletionsStatsView return correct CSV data.
        """
        mock_fetch_and_cache_enrollments_data.return_value = enrollments_dataframe()
        params = {
            'start_date': '2020-01-01',
            'end_date': '2025-08-09',
            'calculation': 'Running Total',
            'granularity': 'Daily',
            'response_type': 'csv',
            'chart_type': chart_type,
        }
        response = self.client.get(self.url, params)
        assert response.status_code == status.HTTP_200_OK
        assert response["Content-Type"] == "text/csv"
        # verify the response content
        assert response.content == COMPLETIONS_STATS_CSVS[chart_type]


@ddt.ddt
class TestCompletionstAPI(JWTTestMixin, APITransactionTestCase):
    """Tests for EnterrpiseAdminCompletionsView."""

    def setUp(self):
        """
        Setup method.
        """
        super().setUp()
        self.user = UserFactory(is_staff=True)
        role, __ = EnterpriseDataFeatureRole.objects.get_or_create(
            name=ENTERPRISE_DATA_ADMIN_ROLE
        )
        self.role_assignment = EnterpriseDataRoleAssignment.objects.create(
            role=role, user=self.user
        )
        self.client.force_authenticate(user=self.user)

        self.enterprise_id = "ee5e6b3a-069a-4947-bb8d-d2dbc323396c"
        self.set_jwt_cookie()

        self.url = reverse(
            "v1:enterprise-admin-analytics-completions",
            kwargs={"enterprise_id": self.enterprise_id},
        )

        fetch_max_enrollment_datetime_patcher = patch(
            'enterprise_data.api.v1.views.enterprise_completions.fetch_max_enrollment_datetime',
            return_value=datetime.now()
        )

        fetch_max_enrollment_datetime_patcher.start()
        self.addCleanup(fetch_max_enrollment_datetime_patcher.stop)

    def verify_enrollment_data(self, results, results_count):
        """Verify the received enrollment data."""
        attrs = [
            "email",
            "course_title",
            "course_subject",
            "passed_date",
        ]

        assert len(results) == results_count

        filtered_data = []
        for enrollment in ENROLLMENTS:
            for result in results:
                if enrollment["email"] == result["email"]:
                    filtered_data.append({attr: enrollment[attr] for attr in attrs})
                    break
        received_data = sorted(results, key=lambda x: x["email"])
        expected_data = sorted(filtered_data, key=lambda x: x["email"])
        assert received_data == expected_data

    @patch(
        "enterprise_data.api.v1.views.enterprise_completions.fetch_and_cache_enrollments_data"
    )
    def test_get(self, mock_fetch_and_cache_enrollments_data):
        """
        Test the GET method for the EnterrpiseAdminCompletionsView works.
        """
        mock_fetch_and_cache_enrollments_data.return_value = enrollments_dataframe()

        response = self.client.get(self.url, {"page": 1, "page_size": 1})
        assert response.status_code == status.HTTP_200_OK
        data = response.json()
        assert data["next"] == f"http://testserver{self.url}?page=2&page_size=1"
        assert data["previous"] is None
        assert data["current_page"] == 1
        assert data["num_pages"] == 2
        assert data["count"] == 2
        self.verify_enrollment_data(data["results"], 1)



"""
Views for enterprise admin api v1.
"""
from datetime import datetime, timedelta

from edx_rbac.decorators import permission_required
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from rest_framework import filters, viewsets
from rest_framework.response import Response
from rest_framework.status import HTTP_200_OK, HTTP_404_NOT_FOUND
from rest_framework.views import APIView

from django.http import HttpResponse

from enterprise_data.admin_analytics.data_loaders import fetch_max_enrollment_datetime
from enterprise_data.admin_analytics.database.tables import FactEngagementAdminDashTable, FactEnrollmentAdminDashTable
from enterprise_data.admin_analytics.utils import (
    ChartType,
    fetch_and_cache_enrollments_data,
    fetch_and_cache_skills_data,
    get_skills_chart_data,
    get_top_skills_csv_data,
)
from enterprise_data.api.v1 import serializers
from enterprise_data.models import (
    EnterpriseAdminLearnerProgress,
    EnterpriseAdminSummarizeInsights,
    EnterpriseExecEdLCModulePerformance,
)
from enterprise_data.utils import timer

from .base import EnterpriseViewSetMixin


class EnterpriseAdminInsightsView(APIView):
    """
    API for getting the enterprise admin insights.
    """

    authentication_classes = (JwtAuthentication,)
    http_method_names = ["get"]

    @permission_required(
        "can_access_enterprise", fn=lambda request, enterprise_id: enterprise_id
    )
    def get(self, request, enterprise_id):
        """
        HTTP GET endpoint to retrieve the enterprise admin insights
        """
        response_data = {}
        learner_progress = {}
        learner_engagement = {}

        try:
            learner_progress = EnterpriseAdminLearnerProgress.objects.get(
                enterprise_customer_uuid=enterprise_id
            )
            learner_progress = serializers.EnterpriseAdminLearnerProgressSerializer(
                learner_progress
            ).data
            response_data["learner_progress"] = learner_progress
        except EnterpriseAdminLearnerProgress.DoesNotExist:
            pass

        try:
            learner_engagement = EnterpriseAdminSummarizeInsights.objects.get(
                enterprise_customer_uuid=enterprise_id
            )
            learner_engagement = serializers.EnterpriseAdminSummarizeInsightsSerializer(
                learner_engagement
            ).data
            response_data["learner_engagement"] = learner_engagement
        except EnterpriseAdminSummarizeInsights.DoesNotExist:
            pass

        status = HTTP_200_OK
        if learner_progress == {} and learner_engagement == {}:
            status = HTTP_404_NOT_FOUND

        return Response(data=response_data, status=status)


class EnterpriseAdminAnalyticsAggregatesView(APIView):
    """
    API for getting the enterprise admin analytics aggregates.
    """

    authentication_classes = (JwtAuthentication,)
    http_method_names = ["get"]

    @permission_required(
        "can_access_enterprise", fn=lambda request, enterprise_id: enterprise_id
    )
    def get(self, request, enterprise_id):
        """
        HTTP GET endpoint to retrieve the enterprise admin aggregate data.
        """
        # Validate the enterprise_id
        enterprise_id = enterprise_id.replace('-', '')
        serializer = serializers.AdminAnalyticsAggregatesQueryParamsSerializer(
            data=request.GET
        )
        serializer.is_valid(raise_exception=True)

        last_updated_at = fetch_max_enrollment_datetime()
        min_enrollment_date, max_enrollment_date = FactEnrollmentAdminDashTable().get_enrollment_date_range(
            enterprise_id,
        )

        start_date = serializer.data.get(
            'start_date', min_enrollment_date
        )
        end_date = serializer.data.get('end_date', datetime.today())

        enrolls, courses = FactEnrollmentAdminDashTable().get_enrollment_and_course_count(
            enterprise_id, start_date, end_date,
        )
        completions = FactEnrollmentAdminDashTable().get_completion_count(
            enterprise_id, start_date, end_date,
        )
        hours, sessions = FactEngagementAdminDashTable().get_learning_hours_and_daily_sessions(
            enterprise_id, start_date, end_date,
        )

        return Response(
            data={
                'enrolls': enrolls,
                'courses': courses,
                'completions': completions,
                'hours': hours,
                'sessions': sessions,
                'last_updated_at': last_updated_at.date() if last_updated_at else None,
                'min_enrollment_date': min_enrollment_date,
                'max_enrollment_date': max_enrollment_date,
            },
            status=HTTP_200_OK,
        )


class EnterpriseAdminAnalyticsSkillsView(APIView):
    """
    API for getting the enterprise admin analytics skills data.
    """
    authentication_classes = (JwtAuthentication,)
    http_method_names = ['get']

    @permission_required(
        'can_access_enterprise', fn=lambda request, enterprise_id: enterprise_id
    )
    def get(self, request, enterprise_id):
        """HTTP GET endpoint to retrieve the enterprise admin skills aggregated data.

        Args:
            request (HttpRequest): request object
            enterprise_id (str): UUID of the enterprise customer

        Returns:
            response(HttpResponse): response object
        """
        serializer = serializers.AdminAnalyticsAggregatesQueryParamsSerializer(
            data=request.GET
        )
        serializer.is_valid(raise_exception=True)
        last_updated_at = fetch_max_enrollment_datetime()
        cache_expiry = (
            last_updated_at + timedelta(days=1) if last_updated_at else datetime.now()
        )

        enrollment = fetch_and_cache_enrollments_data(
            enterprise_id, cache_expiry
        ).copy()

        start_date = serializer.data.get('start_date', enrollment.enterprise_enrollment_date.min())
        end_date = serializer.data.get('end_date', datetime.now())

        skills = fetch_and_cache_skills_data(enterprise_id, cache_expiry).copy()

        if serializer.data.get('response_type') == 'csv':
            csv_data = get_top_skills_csv_data(skills, start_date, end_date)
            response = HttpResponse(content_type='text/csv')
            filename = f"Skills by Enrollment and Completion, {start_date} - {end_date}.csv"
            response['Content-Disposition'] = f'attachment; filename="{filename}"'
            response['Access-Control-Expose-Headers'] = 'Content-Disposition'
            csv_data.to_csv(path_or_buf=response, index=False)
            return response

        with timer('skills_all_charts_data'):
            top_skills = get_skills_chart_data(
                chart_type=ChartType.BUBBLE,
                start_date=start_date,
                end_date=end_date,
                skills=skills,
            )
            top_skills_enrollments = get_skills_chart_data(
                chart_type=ChartType.TOP_SKILLS_ENROLLMENT,
                start_date=start_date,
                end_date=end_date,
                skills=skills,
            )
            top_skills_by_completions = get_skills_chart_data(
                chart_type=ChartType.TOP_SKILLS_COMPLETION,
                start_date=start_date,
                end_date=end_date,
                skills=skills,
            )

        response_data = {
            "top_skills": top_skills.to_dict(orient="records"),
            "top_skills_by_enrollments": top_skills_enrollments.to_dict(
                orient="records"
            ),
            "top_skills_by_completions": top_skills_by_completions.to_dict(
                orient="records"
            ),
        }

        return Response(data=response_data, status=HTTP_200_OK)


class EnterpriseExecEdLCModulePerformanceViewSet(EnterpriseViewSetMixin, viewsets.ReadOnlyModelViewSet):
    """
    View to for getting enterprise exec ed learner module performance records.
    """
    serializer_class = serializers.EnterpriseExecEdLCModulePerformanceSerializer
    filter_backends = (filters.OrderingFilter, filters.SearchFilter)
    ordering_fields = '__all__'
    ordering = ('last_access',)
    search_fields = (
        'username',
        'course_name'
    )

    def get_queryset(self):
        """
        Return the queryset of EnterpriseExecEdLCModulePerformance objects.
        """
        return EnterpriseExecEdLCModulePerformance.objects.filter(
            enterprise_customer_uuid=self.kwargs['enterprise_id'],
        )
